""" Write here the imports"""

from mooda.access.egim import EGIM
from mooda.access.pangaea import Pangaea
from mooda.access.licor import Licor
from mooda.access.hobo import Hobo
from mooda.access.seaandsun import SeaAndSun
from mooda.access.seabird import SeaBird



"""
Module to open data from the txt files of LICOR datalogs
Initial work: Carlos Rodero, https://github.com/Carlos-Rodero/licor-python
"""
import itertools
import pandas as pd
from mooda import WaterFrame


class Licor:
    """
    Class to import Licor data from a txt file.
    It just contains a static method: from_text_to_waterframe(...)
    """

    @staticmethod
    def from_txt_to_waterframe(sensor_model, path, qc_tests=False):
        """
        Parameters
        ----------
            model: str
                Model of the sensor.
            path: str
                Path of the txt file.
            qc_tests: bool (optional)
                It indicates if QC test should be passed.

        Returns
        -------
            wf: WaterFrame
        """
        # Creation of a WaterFrame
        wf = WaterFrame()
        metadata = {}

        # The arguments of pandas.read_csv could change depending on the
        # source.
        if sensor_model == "LI-192":
            # lines to skip (int) at the start of the file.
            skiprows = 6
            # format_time = '%d/%m/%Y %H:%M:%S'

            # Add metadata info
            with open(path, encoding='utf-8-sig') as txtfile:
                # Read 2 first lines from csv
                for row in itertools.islice(txtfile, 1, 6):
                    # Delete return carriage from row
                    row = ''.join(row.splitlines())
                    # Split row by "\t" and remove empty strings returned in
                    # split()
                    parts = row.split(":")
                    if "Timestamp" not in parts[0]:
                        metadata[parts[0]] = parts[1].strip()

            # Load metadata to waterframe
            wf.metadata = metadata

            # Load data from table into a DataFrame
            df = pd.read_table(path, delim_whitespace=True, skiprows=skiprows,
                               low_memory=False)
            df['Ns'] = pd.to_numeric(df['Nanoseconds'], errors='coerce')/1000000000
            df['Input1'] = pd.to_numeric(df['Input1'], errors='coerce')

            # Create the time index. Convert "Nanoseconds" to seconds, add to
            # column "Seconds"
            # and convert to datetime
            df['Seconds_total'] = df['Ns'] + pd.to_numeric(df['Seconds'],
                                                           errors='coerce')
            df['TIME'] = pd.to_datetime(df['Seconds_total'], unit='s',
                                        errors='coerce')
            df.set_index(df['TIME'], inplace=True)
            df.drop(['DATAH', 'Record', 'Seconds', 'Nanoseconds', 'Ns',
                     'Seconds_total', 'TIME', 'MULT_1', 'CHK'], inplace=True,
                    axis=1)

            # Add DataFrame into the WaterFrame
            wf.data = df.copy()

            # Change parameter names and add QC columns
            for key in wf.data.keys():
                if key == "Input1":
                    wf.data.rename(columns={"Input1": "PPFD"}, inplace=True)
                    wf.data["PPFD_QC"] = 0
                    wf.meaning['PPFD'] = {"long_name":
                                          "Photosynthetic Photon Flux Density",
                                          "units": "µMol/M^2S"}

            # resample to seconds
            wf.resample('S')

            # Creation of QC Flags following OceanSites recomendation
            if qc_tests:
                # Reset QC Flags to 0
                wf.reset_flag(flag=0)
                # Flat test
                wf.flat_test(window=0, flag=4)
                # Spike test
                wf.spike_test(window=0, threshold=3, flag=4)
                # Range test
                wf.range_test(flag=4)
                # Change flags from 0 to 1
                wf.flag2flag(original_flag=0, translated_flag=1)

        else:
            raise ValueError("model not found or implemented")

        return wf



""" Module to access to the Data Management Portal of EMSO Eric"""
# pylint: disable=C0302
import ast
import datetime
from io import StringIO
import warnings
import requests
import pandas as pd
import xarray as xr
from mooda import WaterFrame


class EGIM:
    """Class to download EGIM data from the EMSODEV DMP."""

    METADATA_00001 = {'site_code': 'OBSEA',
                      'platform_code': 'prototype1',
                      'data_mode': 'D',
                      'title': ('Test dataset from EMSODEV shallow water test '
                                'side'),
                      'summary': ('Test dataset from EMSODEV shallow water '
                                  'test side in the Mediterranean Sea'),
                      'naming_authority': 'UPC-CSIC',
                      'source': 'subsurface mooring',
                      'institution': 'EMSO',
                      'project': 'EMSODEV',
                      'keywords_vocabulary': 'GCMD Science Keywords',
                      'comment': 'Test data',
                      'geospatial_lat_min': '41.182',
                      'geospatial_lat_max': '41.182',
                      'geospatial_lat_units': 'degree_north',
                      'geospatial_lon_min': '1.752',
                      'geospatial_lon_max': '1.752',
                      'geospatial_lon_units': 'degree_east',
                      'geospatial_vertical_min': '17.841',
                      'geospatial_vertical_max': '19.879',
                      'geospatial_vertical_positive': 'down',
                      'geospatial_vertical_units': 'meter',
                      'time_coverage_start': '2017-01-27T09:28:57Z',
                      'time_coverage_end': '2017-04-24T07:11:24Z',
                      'time_coverage_duration': '',
                      'time_coverage_resolution': 'PT30S',
                      'cmd_data_type': 'Station',
                      'data_type': 'OceanSITES time-series data',
                      'format_version': '1.3',
                      'publisher_name': '',
                      'publisher_email': '',
                      'publisher_url': '',
                      'references': ('http://www.oceansites.org, '
                                     'http://emso.eu'),
                      'data_assembly_center': '',
                      'update_interval': 'void',
                      'license': ('Follows CLIVAR (Climate Variability and '
                                  'Predictability) standards, cf. '
                                  'http://www.clivar.org/data/data_policy.php '
                                  'Data available free of charge. User '
                                  'assumes all risk for use of data. Uer must '
                                  'display citation in any publication or '
                                  'product using data. User must contact PI '
                                  'prior to any commercial use of data.'),
                      'citation': ('These data were collected and made freely '
                                   'available by the EMSODEV project.'),
                      'acknowledgement': ('This work benefited from the H2020 '
                                          'INFRADEV-3-2015 EMSODEV Project '
                                          'n°676555.'),
                      'date_created':
                      datetime.datetime.utcnow().isoformat()[:-6]+'Z',
                      'date_modified':
                      datetime.datetime.utcnow().isoformat()[:-6]+'Z',
                      'history': 'Data collected and processed with MOODA',
                      'processing_level': 'Ranges applied, bad data flagged',
                      'QC_indicator': 'excellent',
                      'contributor_name': 'EMSODEV project',
                      'contributor_role': 'consortium',
                      'contributor_email': ''}

    METADATA_AZORES = {'site_code': 'Azores',
                       'platform_code': 'EMSO-Azores',
                       'data_mode': 'D',
                       'title': 'Data from EGIM in the EMSO Site Azores',
                       'summary': ('This dataset contains pressure data '
                                   'acquired between July 2017 and August '
                                   '2018 on EMSO-Azores observatory by the '
                                   'EGIM. The pressure is one of the 7 core '
                                   'parameters monitored by the EGIM, EMSO '
                                   'Generic Instrumental Module. The EGIM '
                                   'prototype was deployed at Lucky Strike '
                                   'hydrothermal vent site, 25 m south west of'
                                   ' the active edifice Tour Eiffel, to '
                                   'monitor local hydrodynamic variability and'
                                   ' complement the data obtained by the '
                                   'numerous sensors set on this site: '
                                   'oceanographic mooring deployed south of '
                                   'the vent field, the multidisciplinary '
                                   'Seamon East node, autonomous current '
                                   'meters, array of temperature probes.'),
                       'naming_authority': '',
                       'source': 'subsurface mooring',
                       'institution': 'Ifremer, CNRS, IPGP',
                       'project': 'EMSODEV',
                       'keywords_vocabulary': 'GCMD Science Keywords',
                       'comment': 'Test data',
                       'geospatial_lat_min': '37.289433',
                       'geospatial_lat_max': '37.289433',
                       'geospatial_lat_units': 'degree_north',
                       'geospatial_lon_min': '-32.277567',
                       'geospatial_lon_mX': '-32.277567',
                       'geospatial_lon_units': 'degree_east',
                       'geospatial_vertical_min': '17.841',
                       'geospatial_vertical_max': '19.879',
                       'geospatial_vertical_positive': 'down',
                       'geospatial_vertical_units': 'meter',
                       'time_coverage_start': '2017-07-01T00:00:00Z',
                       'time_coverage_end': '2018-08-31T00:00:00Z',
                       'time_coverage_duration': '',
                       'time_coverage_resolution': 'PT30S',
                       'cmd_data_type': 'Station',
                       'data_type': 'OceanSITES time-series data',
                       'format_version': '1.3',
                       'publisher_name': '',
                       'publisher_email': '',
                       'publisher_url': '',
                       'references': ('http://www.oceansites.org, '
                                      'http://emso.eu'),
                       'data_assembly_center': '',
                       'update_interval': 'void',
                       'license': ('https://creativecommons.org/'
                                   'licenses/by/4.0/'),
                       'citation': ('These data were collected and made freely'
                                    ' available by the EMSODEV project.'),
                       'acknowledgement': ('This work benefited from the H2020'
                                           ' INFRADEV-3-2015 EMSODEV Project '
                                           'n°676555.'),
                       'date_created':
                       datetime.datetime.utcnow().isoformat()[:-6]+'Z',
                       'date_modified':
                       datetime.datetime.utcnow().isoformat()[:-6]+'Z',
                       'history': 'Data collected and processed with MOODA',
                       'processing_level': 'Ranges applied, bad data flagged',
                       'QC_indicator': 'excellent',
                       'contributor_name': 'EMSODEV project',
                       'contributor_role': 'consortium',
                       'contributor_email': ''}

    METADATA_ADCP_21582 = {'sensor_model': 'TELEDYNE RDI Workhorse monitor',
                           'sensor_manufactured': 'TELEDYNE RD INSTRUMENTS',
                           'sensor_reference': 'Workhorse_ADCP_21582',
                           'sensor_serial_number': '21582',
                           'sensor_mount':
                           'mounted_on_seafloor_structure_riser',
                           'sensor_orientation': 'downward',
                           }

    METADATA_icListen_1636 = {'sensor_model': 'OceanSonics icListen SB60L-ETH',
                              'sensor_manufactured': 'Ocean Sonics',
                              'sensor_reference': 'icListen-1636',
                              'sensor_serial_number': '1636',
                              'sensor_mount':
                              'mounted_on_seafloor_structure_riser',
                              'sensor_orientation': 'downward',
                              }

    METADATA_TEMP_SBE37 = {'standard_name': 'sea_water_temperature',
                           'units': 'degree_Celsius',
                           'coordinates': 'TIME DEPTH LATITUDE LONGITUDE',
                           'long_name': ' Temperature of the water column',
                           'QC_indicator': 'Good data',
                           'processing_level': ('Ranges applied, bad data '
                                                'flagged'),
                           'valid_min': '-5.f',
                           'valid_max': '45.f',
                           'ancillary_variables': 'TEMP_QC',
                           'history': '',
                           'uncertainty': '0.002f',
                           'accuracy': '0.002f',
                           'precision': '0.002f',
                           'resolution': '0.002f',
                           'cell_methods': ('TIME: mean DEPTH: point '
                                            'LATITUDE: point LONGITUDE: '
                                            'point'),
                           'DM_indicator': 'D',
                           'reference_scale': 'ITS-90',
                           'sensor_model': 'SBE37-SIP-P7000-RS232',
                           'sensor_manufactured': 'Sea-Bird Scientific',
                           'sensor_reference': '37-14998',
                           'sensor_serial_number': '14998',
                           'sensor_mount':
                           'mounted_on_seafloor_structure_riser',
                           'sensor_orientation': 'downward',
                           }

    METADATA_TEMP_AADI4381 = {'standard_name': 'sea_water_temperature',
                              'units': 'degree_Celsius',
                              'coordinates': 'TIME DEPTH LATITUDE LONGITUDE',
                              'long_name': 'Temperature of the water column',
                              'QC_indicator': 'Good data',
                              'processing_level': ('Ranges applied, bad data '
                                                   'flagged'),
                              'valid_min': '-5.f',
                              'valid_max': '40.f',
                              'ancillary_variables': 'TEMP_QC',
                              'history': '',
                              'uncertainty': '0.03f',
                              'accuracy': '0.03f',
                              'precision': '0.03f',
                              'resolution': '0.01f',
                              'cell_methods': ('TIME: mean DEPTH: point '
                                               'LATITUDE: point LONGITUDE: '
                                               'point'),
                              'DM_indicator': 'D',
                              'reference_scale': 'ITS-90',
                              'sensor_model': 'AADI-3005214831 DW4831',
                              'sensor_manufactured': ('Aanderaa Data '
                                                      'Instruments AS'),
                              'sensor_reference': '4381-606',
                              'sensor_serial_number': '606',
                              'sensor_mount':
                              'mounted_on_seafloor_structure_riser',
                              'sensor_orientation': 'downward',
                              }

    METADATA_TEMP_QC = {'long_name': 'quality flag for sea water temperature',
                        'flag_values': '0, 1, 2, 3, 4, 7, 8, 9',
                        'flag_meanings': ('unknown good_data '
                                          'probably_good_data '
                                          'potentially_correctable_bad_data '
                                          'bad_data nominal_value '
                                          'interpolated_value missing_value'),
                        }

    METADATA_PSAL_SBE37 = {'standard_name': 'sea_water_practical_salinity',
                           'units': 'PSU',
                           'coordinates': 'TIME DEPTH LATITUDE LONGITUDE',
                           'long_name': 'Salinity of the water column',
                           'QC_indicator': 'Good data',
                           'processing_level': ('Ranges applied, bad data '
                                                'flagged'),
                           'valid_min': '33.f',
                           'valid_max': '37.f',
                           'ancillary_variables': 'PSAL_QC',
                           'history': '',
                           'uncertainty': '',
                           'accuracy': '',
                           'precision': '',
                           'resolution': '',
                           'cell_methods': ('TIME: mean DEPTH: point '
                                            'LATITUDE: point LONGITUDE: '
                                            'point'),
                           'DM_indicator': 'D',
                           'reference_scale': '',
                           'sensor_model': 'SBE37-SIP-P7000-RS232',
                           'sensor_manufactured': 'Sea-Bird Scientific',
                           'sensor_reference': '37-14998',
                           'sensor_serial_number': '14998',
                           'sensor_mount':
                           'mounted_on_seafloor_structure_riser',
                           'sensor_orientation': 'downward',
                           }

    METADATA_PSAL_QC = {'long_name': 'quality flag for sea water practical salinity',
                        'flag_values': '0, 1, 2, 3, 4, 7, 8, 9',
                        'flag_meanings': ('unknown good_data '
                                          'probably_good_data '
                                          'potentially_correctable_bad_data '
                                          'bad_data nominal_value '
                                          'interpolated_value missing_value'),
                        }

    METADATA_CNDC_SBE37 = {'standard_name':
                           'sea_water_electrical_conductivity',
                           'units': 'S/m',
                           'coordinates': 'TIME DEPTH LATITUDE LONGITUDE',
                           'long_name': 'Electrical conductivity of the water column',
                           'QC_indicator': 'Good data',
                           'processing_level': 'Ranges applied, bad data flagged',
                           'valid_min': '0.f',
                           'valid_max': '7.f',
                           'ancillary_variables': 'CNDC_QC',
                           'history': '',
                           'uncertainty': '0.0003f',
                           'accuracy': '0.0003f',
                           'precision': '0.0003f',
                           'resolution': '0.0003f',
                           'cell_methods': ('TIME: mean DEPTH: point '
                                            'LATITUDE: point LONGITUDE: '
                                            'point'),
                           'DM_indicator': 'D',
                           'reference_scale': '',
                           'sensor_model': 'SBE37-SIP-P7000-RS232',
                           'sensor_manufactured': 'Sea-Bird Scientific',
                           'sensor_reference': '37-14998',
                           'sensor_serial_number': '14998',
                           'sensor_mount':
                           'mounted_on_seafloor_structure_riser',
                           'sensor_orientation': 'downward',
                           }

    METADATA_CNDC_QC = {'long_name': ('quality flag for sea water electrical '
                                      'conductivity'),
                        'flag_values': '0, 1, 2, 3, 4, 7, 8, 9',
                        'flag_meanings': ('unknown good_data '
                                          'probably_good_data '
                                          'potentially_correctable_bad_data '
                                          'bad_data nominal_value '
                                          'interpolated_value missing_value'),
                        }

    METADATA_MPMN_SBE37 = {'standard_name': 'depth',
                           'units': 'meters',
                           'coordinates': 'TIME DEPTH LATITUDE LONGITUDE',
                           'long_name': 'Moored instrument depth',
                           'QC_indicator': 'Good data',
                           'processing_level': ('Ranges applied, bad data '
                                                'flagged'),
                           'valid_min': '0.f',
                           'valid_max': '7000.f',
                           'ancillary_variables': 'MPMN_QC',
                           'history': '',
                           'uncertainty': '',
                           'accuracy': '',
                           'precision': '',
                           'resolution': '',
                           'cell_methods': ('TIME: mean DEPTH: point '
                                            'LATITUDE: point LONGITUDE: '
                                            'point'),
                           'DM_indicator': 'D',
                           'reference_scale': '',
                           'sensor_model': 'SBE37-SIP-P7000-RS232',
                           'sensor_manufactured': 'Sea-Bird Scientific',
                           'sensor_reference': '37-14998',
                           'sensor_serial_number': '14998',
                           'sensor_mount':
                           'mounted_on_seafloor_structure_riser',
                           'sensor_orientation': 'downward',
                           }

    METADATA_MPMN_QC = {'long_name': 'quality flag for instrument depth',
                        'flag_values': '0, 1, 2, 3, 4, 7, 8, 9',
                        'flag_meanings': ('unknown good_data '
                                          'probably_good_data '
                                          'potentially_correctable_bad_data '
                                          'bad_data nominal_value '
                                          'interpolated_value missing_value'),
                        }

    METADATA_SVEL_SBE37 = {'standard_name': 'sea_water_sound_velocity',
                           'units': 'meters/second',
                           'coordinates': 'TIME DEPTH LATITUDE LONGITUDE',
                           'long_name': 'Sound velocity of the water column',
                           'QC_indicator': 'Good data',
                           'processing_level': 'Ranges applied, bad data flagged',
                           'valid_min': '',
                           'valid_max': '',
                           'ancillary_variables': 'SVEL_QC',
                           'history': '',
                           'uncertainty': '',
                           'accuracy': '',
                           'precision': '',
                           'resolution': '',
                           'cell_methods': ('TIME: mean DEPTH: point '
                                            'LATITUDE: point LONGITUDE: '
                                            'point'),
                           'DM_indicator': 'D',
                           'reference_scale': '',
                           'sensor_model': 'SBE37-SIP-P7000-RS232',
                           'sensor_manufactured': 'Sea-Bird Scientific',
                           'sensor_reference': '37-14998',
                           'sensor_serial_number': '14998',
                           'sensor_mount':
                           'mounted_on_seafloor_structure_riser',
                           'sensor_orientation': 'downward',
                           }

    METADATA_SVEL_QC = {'long_name': ('quality flag for sea water sound '
                                      'velocity'),
                        'flag_values': '0, 1, 2, 3, 4, 7, 8, 9',
                        'flag_meanings': ('unknown good_data '
                                          'probably_good_data '
                                          'potentially_correctable_bad_data '
                                          'bad_data nominal_value '
                                          'interpolated_value missing_value'),
                        }

    METADATA_PRES_SBE54 = {'standard_name': 'sea_water_pressure',
                           'units': 'psia',
                           'coordinates': 'TIME DEPTH LATITUDE LONGITUDE',
                           'long_name': 'Pressure of water',
                           'QC_indicator': 'Good data',
                           'processing_level': ('Ranges applied, bad data '
                                                'flagged'),
                           'valid_min': '0.f',
                           'valid_max': '10000.f',
                           'ancillary_variables': 'PRES_QC',
                           'history': '',
                           'uncertainty': '0.0011f',
                           'accuracy': '0.0011f',
                           'precision': '0.0011f',
                           'resolution': '0.0011f',
                           'cell_methods': ('TIME: mean DEPTH: point '
                                            'LATITUDE: point LONGITUDE: '
                                            'point'),
                           'DM_indicator': 'D',
                           'reference_scale': '',
                           'sensor_model': 'SBE54-0049',
                           'sensor_manufactured': 'Sea-Bird Scientific',
                           'sensor_reference': '54-0049',
                           'sensor_serial_number': '0049',
                           'sensor_mount':
                           'mounted_on_seafloor_structure_riser',
                           'sensor_orientation': 'downward',
                           }

    METADATA_PRES_QC = {'long_name': 'quality flag for sea water pressure',
                        'flag_values': '0, 1, 2, 3, 4, 7, 8, 9',
                        'flag_meanings': ('unknown good_data '
                                          'probably_good_data '
                                          'potentially_correctable_bad_data '
                                          'bad_data nominal_value '
                                          'interpolated_value missing_value'),
                        }

    METADATA_TUR4_NTURTD = {'standard_name': 'turbidity',
                            'units': 'NTU',
                            'coordinates': 'TIME DEPTH LATITUDE LONGITUDE',
                            'long_name': 'Turbidity of water',
                            'QC_indicator': 'Good data',
                            'processing_level': ('Ranges applied, bad data '
                                                 'flagged'),
                            'valid_min': '',
                            'valid_max': '',
                            'ancillary_variables': 'TUR4_QC',
                            'history': '',
                            'uncertainty': '',
                            'accuracy': '',
                            'precision': '',
                            'resolution': '',
                            'cell_methods': ('TIME: mean DEPTH: point '
                                             'LATITUDE: point LONGITUDE: '
                                             'point'),
                            'DM_indicator': 'D',
                            'reference_scale': '',
                            'sensor_model': 'WETlabs ECO NTUrtd',
                            'sensor_manufactured': 'Sea-Bird Scientific',
                            'sensor_reference': 'NTURTD-648',
                            'sensor_serial_number': '648',
                            'sensor_mount':
                            'mounted_on_seafloor_structure_riser',
                            'sensor_orientation': 'horizontal',
                            }

    METADATA_TUR4_QC = {'long_name': 'quality flag for sea water turbidity',
                        'flag_values': '0, 1, 2, 3, 4, 7, 8, 9',
                        'flag_meanings': ('unknown good_data '
                                          'probably_good_data '
                                          'potentially_correctable_bad_data '
                                          'bad_data nominal_value '
                                          'interpolated_value missing_value'),
                        }

    METADATA_OSAT_AADI4381 = {'standard_name': 'oxygen_saturation',
                              'units': '%',
                              'coordinates': 'TIME DEPTH LATITUDE LONGITUDE',
                              'long_name': 'oxygen saturation of water',
                              'QC_indicator': 'Good data',
                              'processing_level': ('Ranges applied, bad data '
                                                   'flagged'),
                              'valid_min': '0.f',
                              'valid_max': '100.f',
                              'ancillary_variables': 'OSAT_QC',
                              'history': '',
                              'uncertainty': '5f',
                              'accuracy': '5f',
                              'precision': '5f',
                              'resolution': '0.4f',
                              'cell_methods': ('TIME: mean DEPTH: point '
                                               'LATITUDE: point LONGITUDE: '
                                               'point'),
                              'DM_indicator': 'D',
                              'reference_scale': '',
                              'sensor_model': 'AADI-3005214831 DW4831',
                              'sensor_manufactured': ('Aanderaa Data '
                                                      'Instruments AS'),
                              'sensor_reference': '4381-606',
                              'sensor_serial_number': '606',
                              'sensor_mount':
                              'mounted_on_seafloor_structure_riser',
                              'sensor_orientation': 'downward',
                              }

    METADATA_OSAT_QC = {'long_name': 'quality flag for oxygen saturation',
                        'flag_values': '0, 1, 2, 3, 4, 7, 8, 9',
                        'flag_meanings': ('unknown good_data '
                                          'probably_good_data '
                                          'potentially_correctable_bad_data '
                                          'bad_data nominal_value '
                                          'interpolated_value missing_value'),
                        }

    METADATA_DOX2_AADI4381 = {'standard_name': 'dissolbed_oxygen',
                              'units': 'microMols / liter',
                              'coordinates': 'TIME DEPTH LATITUDE LONGITUDE',
                              'long_name': 'moles of oxygen per unit mass',
                              'QC_indicator': 'Good data',
                              'processing_level': ('Ranges applied, bad data '
                                                   'flagged'),
                              'valid_min': '0.f',
                              'valid_max': '500.f',
                              'ancillary_variables': 'DOX2_QC',
                              'history': '',
                              'uncertainty': '8f',
                              'accuracy': '8f',
                              'precision': '8f',
                              'resolution': '2.5f',
                              'cell_methods': ('TIME: mean DEPTH: point '
                                               'LATITUDE: point LONGITUDE: '
                                               'point'),
                              'DM_indicator': 'D',
                              'reference_scale': '',
                              'sensor_model': 'AADI-3005214831 DW4831',
                              'sensor_manufactured': ('Aanderaa Data '
                                                      'Instruments AS'),
                              'sensor_reference': '4381-606',
                              'sensor_serial_number': '606',
                              'sensor_mount':
                              'mounted_on_seafloor_structure_riser',
                              'sensor_orientation': 'downward',
                              }

    METADATA_DOX2_QC = {'long_name': 'quality flag for dissolved oxygen',
                        'flag_values': '0, 1, 2, 3, 4, 7, 8, 9',
                        'flag_meanings': ('unknown good_data '
                                          'probably_good_data '
                                          'potentially_correctable_bad_data '
                                          'bad_data nominal_value '
                                          'interpolated_value missing_value'),
                        }

    def __init__(self, login=None, password=None):
        """
        It creates the instance variables login and password to use the DMP
        API.

        Parameters
        ----------
            login: str
                Login of the EMSODEV DMP API.
            password: str
                Password of the EMSODEV DMP API.
        """
        self.login = login
        self.password = password

    def observatories(self):
        """
        It represents the EGIM observatories accessible through the EMSODEV
        DMP API.

        Returns
        -------
            (statusCode, observatoryList): (int, list of str)
                (Status code answer of the API, list with names of
                observatories).
        """
        try:
            response = requests.get('http://api.emsodev.eu/observatories',
                                    auth=(self.login, self.password))
        except requests.RequestException:
            return None, None
        if response.status_code == 200:
            answer = response.json()
            observatory_list = [observatory['name'] for observatory in answer]
            return response.status_code, observatory_list
        else:
            return response.status_code, None

    def instruments(self, observatory):
        """
        It represents the instruments deployed in an EGIM observatory.

        Parameters
        ----------
            observatory: str
                EGIM observatory name.

        Returns
        -------
            (statusCode, instrumentList): (int, list of dict{"name": "string",
            "sensorLongName": "string", "sensorType": "string",
            "sn": "string"})
                (Status code answer of the API, list with dictionaries of
                available instruments)
        """
        try:
            response = requests.get(
                'http://api.emsodev.eu/observatories/{}/instruments'.format(
                    observatory), auth=(self.login, self.password))
        except requests.RequestException:
            return None, None
        if response.status_code == 200:
            answer = response.json()
            instrument_list = answer['instruments']
            return response.status_code, instrument_list
        else:
            return response.status_code, None

    def metadata(self, observatory, instrument):
        """
        Get EGIM observatory instrument resource.

        Parameters
        ----------
            observatory: str
                EGIM observatory name.
            instrument: str
                Instrument name.

        Return
        ------
            (statusCode, metadataList): (int, list of dict{{"instrumentName":
            "string", "metadataList": [{"metadata": "string","validityDate":
            "string"}]}})
                (Status code answer of the API, list with dictionaries of
                metadata).
        """
        try:
            response = requests.get(
                'http://api.emsodev.eu/observatories/{}/instruments/{}'.format(
                    observatory, instrument), auth=(self.login, self.password))
        except requests.RequestException:
            return None, None
        if response.status_code == 200:
            answer = response.json()
            # The DMP API do not response with a JSON well formated.
            # We are going to fix their error with the following lines.
            metadata_dict = dict(answer['metadataList'][0])
            metadata_dict = metadata_dict['metadata']
            metadata_dict = metadata_dict.replace('{E', '{\'E')
            metadata_dict = metadata_dict.replace('=', '\':\'')
            metadata_dict = metadata_dict.replace(', ', '\', \'')
            metadata_dict = metadata_dict.split(", 'InstrumentPosition'")[0]
            metadata_dict += "}"
            metadata_dict = ast.literal_eval(metadata_dict)
            return response.status_code, metadata_dict
        else:
            return response.status_code, None

    def parameters(self, observatory, instrument):
        """
        Get the list of EGIM parameters for a specific EGIM instrument of an
        EGIM Observatory.

        Parameters
        ----------
            observatory: str
                EGIM observatory name.
            instrument: str
                Instrument name

        Returns
        -------
            (statusCode, parameterList): (int, list of dict{"name": "string",
            "uom": "string"})
                (Status code answer of the API, list of dict of parameters)
        """
        try:
            response = requests.get(
                'http://api.emsodev.eu/observatories/{}/instruments/{}/'
                'parameters'.format(observatory, instrument),
                auth=(self.login, self.password))
        except requests.RequestException:
            return None, None
        if response.status_code == 200:
            answer = response.json()
            parameter_list = answer['parameters']
            return response.status_code, parameter_list
        else:
            return response.status_code, None

    def observation(self, observatory, instrument, parameter, start_date=None,
                    end_date=None, limit=None):
        """
        Gets the time-series of a specific EGIM parameter in a certain time
        range or  the last X (limit) values for an EGIM instrument of an EGIM
        observatory.

        Parameters
        ----------
            observatory: str
                EGIM observatory name.
            instrument: str
                Instrument name.
            parameter: str
                Parameter name.
            start_date: str, optional (start_date = None)
                Beginning date for the time series range. The date format is
                dd/MM/yyyy.
                If the start time is not supplied, we are going to use 'limit'.
            end_date: str
                End date for the time series range. The date format is
                dd/MM/yyyy.
                If the end time is not supplied, the current time will be used.
            limit: str
                The last x-measurements.
        Returns
        -------
            (statusCode, data): (int, list of Dataframe)
                (Status code answer of the API, list with dict of parameters)
        """

        # Query definition
        query = ''
        if limit is None:
            limit = 0
        else:
            limit = int(limit)
        if limit > 0:
            query = 'http://api.emsodev.eu/observatories' + \
                    '/{}/instruments/{}/parameters/{}/limit/{}'.format(
                        observatory, instrument, parameter, limit)
        else:
            if start_date and end_date:
                query = 'http://api.emsodev.eu/observatories' + \
                        '/{}/instruments'.format(observatory) + \
                        '/{}/parameters/{}?startDate={}&endDate={}'.\
                    format(instrument, parameter, start_date,
                           end_date)
            elif start_date:
                query = 'http://api.emsodev.eu/observatories' + \
                        '/{}/instruments/{}/parameters/{}?startDate={}'.\
                        format(observatory, instrument, parameter, start_date)
        try:
            response = requests.get(query, auth=(self.login, self.password))
        except requests.RequestException:
            return None, None
        if response.status_code == 200:
            answer = response.json()
            observations = []
            for observation in answer['observations']:
                observations.append((observation['phenomenonTime'],
                                     observation['value']))
            # Format data
            df = pd.DataFrame({parameter: [x[1] for x in observations],  # pylint: disable: C0103
                               'time': [x[0] for x in observations]})
            # Changing the time values to a datatime
            df['time'] = pd.to_datetime(df['time'], unit='s')
            df.rename(columns={"time": "TIME"}, inplace=True)
            df.set_index('TIME', inplace=True)

            return response.status_code, df
        else:
            return response.status_code, None

    def acoustic_date(self, observatory, instrument):
        """
        Gets the date list of available acoustic files observed by a specific
        EGIM instrument of an EGIM Observatory.

        Parameters
        ----------
        observatory: str
                EGIM observatory name.
            instrument: str
                Instrument name.
        Returns
        -------
            (statusCode, dateList): (int, list of dict{})
                (Status code answer of the API, list with dict of dates)
        """
        try:
            response = requests.get('http://api.emsodev.eu/observatories' +
                                    '/{}/instruments/{}/acousticfiledate'.format(
                                        observatory, instrument),
                                    auth=(self.login, self.password))
        except requests.RequestException:
            return None, None
        if response.status_code == 200:
            answer = response.json()
            dateList = answer['acousticObservationDate']
            return response.status_code, dateList
        else:
            return response.status_code, None

    def acoustic_observation(self, observatory, instrument, date, hour_minute):
        """
        Gets an Acoustic file for a specific EGIM instrument of an EGIM
        Observatory.

        Parameters
        ----------
            observatory: str
                EGIM observatory name.
            instrument: str
                Instrument name.
            date: str
                Date of Acoustic file. The date format is dd/MM/yyyy.
            hour_minute: str
                Hour and Minute of an Acoustic file. The Hour Minute format is
                HHMM.
        Returns
        -------
            (statusCode, text): (int, str)
                (Status code answer of the API, text of the acoustic file)
        """
        try:
            response = requests.get(
                'http://api.emsodev.eu/observatories' +
                '/{}/instruments/{}/acousticfile?date={}&hourMinute={}'.format(
                    observatory, instrument, date, hour_minute), auth=(
                        self.login, self.password))
        except requests.RequestException:
            return None, None
        if response.status_code == 200:

            # Write metadata
            metadata = {}
            lines = response.text.split("\n")
            for i, line in enumerate(lines):
                print(i, line)
                if i in [0, 1, 9, 10, 17, 18]:
                    continue
                if i == 28:
                    break
                parts = line.split("\t")
                metadata[parts[0]] = parts[1]

            # Write data
            data_ = StringIO(response.text.split("Data:")[1])
            data = pd.read_csv(data_, sep='\t')

            # Changing the time values to a datatime
            data['Time'] = pd.to_datetime(data['Time'])
            data.rename(columns={"Time": "TIME"}, inplace=True)
            data.set_index('TIME', inplace=True)

            return response.status_code, data, metadata
        else:
            return response.status_code, None

    @staticmethod
    def to_waterframe(data, metadata):
        """
        It creates a WaterFrame object from the input variables.

        Parameters
        ----------
            data: Pandas DataFrame
                Pandas DataFrame with data without WaterFrame format.
            metadata: dict
                Dictionary with metadata information.
        Returns
        -------
            wf: WaterFrame
                Data and metadata formated in a WaterFrame Object.
        """

        # Delete columns without data
        for key in data.keys():
            if data[key].empty:
                del data[key]

        wf = WaterFrame()
        wf.data = data

        # Change names
        for key in data.keys():
            if key == "depth":
                wf.data.rename(columns={"depth": "MPMN"}, inplace=True)
                wf.data["MPMN_QC"] = 0
                wf.meaning['MPMN'] = {"long_name": "depth", "units": "meters"}
            elif key == "salinity":
                wf.data.rename(columns={"salinity": "PSAL"}, inplace=True)
                wf.data["PSAL_QC"] = 0
                wf.meaning['PSAL'] = {
                    "long_name": "sea_water_practical_salinity",
                    "units": "PSU"}
            elif key == "conductivity":
                wf.data.rename(columns={"conductivity": "CNDC"}, inplace=True)
                wf.data["CNDC_QC"] = 0
                wf.meaning['CNDC'] = {
                    "long_name": "sea_water_electrical_conductivity",
                    "units": "S/m"}
            elif key == "sea_water_temperature":
                wf.data.rename(columns={"sea_water_temperature": "TEMP"},
                               inplace=True)
                wf.data["TEMP_QC"] = 0
                wf.meaning['TEMP'] = {"long_name": "sea_water_temperature",
                                      "units": "degree Celsius"}
            elif key == "sound_velocity":
                wf.data.rename(columns={"sound_velocity": "SVEL"},
                               inplace=True)
                wf.data["SVEL_QC"] = 0
                wf.meaning['SVEL'] = {"long_name": "sound_velocity",
                                      "units": "m/s"}
            elif key == "oxygen_saturation":
                wf.data.rename(columns={"oxygen_saturation": "OSAT"},
                               inplace=True)
                wf.data["OSAT_QC"] = 0
                wf.meaning['OSAT'] = {"long_name": "oxygen_saturation",
                                      "units": "%"}
            elif key == "dissolved_oxygen":
                wf.data.rename(columns={"dissolved_oxygen": "DOX2"},
                               inplace=True)
                wf.data["DOX2_QC"] = 0
                wf.meaning['DOX2'] = {
                    "long_name": "moles_of_oxygen_per_unit_mass",
                    "units": "uM/l"}
            elif key == "turbidity":
                wf.data.rename(columns={"turbidity": "TUR4"}, inplace=True)
                wf.data["TUR4_QC"] = 0
                wf.meaning['TUR4'] = {"long_name": "turbidity", "units": "NTU"}
            elif key == "sea_water_pressure":
                wf.data.rename(columns={"sea_water_pressure": "PRES"},
                               inplace=True)
                wf.data["PRES_QC"] = 0
                wf.meaning['PRES'] = {"long_name": "sea_water_pressure",
                                      "units": "PSI"}
            elif "N_S_sea_water_speed" in key:
                # Obtain the Bin number
                bin_number = key.split("_")[0].split("n")[1]
                wf.data.rename(columns={key: "VCUR{}".format(bin_number)},
                               inplace=True)
                wf.data["VCUR{}_QC".format(bin_number)] = 0
                wf.meaning['VCUR{}'.format(bin_number)] = {
                    "long_name": "northward_sea_water_velocity ",
                    "units": "mm/s"}
            elif "E_W_sea_water_speed" in key:
                # Obtain the Bin number
                bin_number = key.split("_")[0].split("n")[1]
                wf.data.rename(columns={key: "UCUR{}".format(bin_number)},
                               inplace=True)
                wf.data["UCUR{}_QC".format(bin_number)] = 0
                wf.meaning['UCUR{}'.format(bin_number)] = {
                    "long_name": "eastward_sea_water_velocity",
                    "units": "mm/s"}
            elif "error_sea_water_speed" in key:
                # Obtain the Bin number
                bin_number = key.split("_")[0].split("n")[1]
                wf.data.rename(columns={key: "error_sea_water_speed{}".format(
                    bin_number)}, inplace=True)
                wf.data["error_sea_water_speed{}_QC".format(bin_number)] = 0
                wf.meaning['error_sea_water_speed{}'.format(bin_number)] = {
                    "long_name": "error_sea_water_speed",
                    "units": "mm/s"}
            elif "vert_sea_water_speed" in key:
                # Obtain the Bin number
                bin_number = key.split("_")[0].split("n")[1]
                wf.data.rename(columns={key: "VCSP{}".format(bin_number)},
                               inplace=True)
                wf.data["VCSP{}_QC".format(bin_number)] = 0
                wf.meaning['VCSP{}'.format(bin_number)] = {
                    "long_name": "vert_sea_water_speed",
                    "units": "mm/s"}
            elif key == "pitch":
                wf.data.rename(columns={"pitch": "pitch"}, inplace=True)
                wf.data["pitch_QC"] = 0
                wf.meaning['pitch'] = {"long_name": "pitch",
                                       "units": "degrees"}
            elif key == "heading_of_device":
                wf.data.rename(columns={"heading_of_device": "heading"},
                               inplace=True)
                wf.data["heading_QC"] = 0
                wf.meaning['heading'] = {"long_name": "heading_of_device",
                                         "units": "degrees"}
            elif key == "roll":
                wf.data.rename(columns={"roll": "roll"}, inplace=True)
                wf.data["roll_QC"] = 0
                wf.meaning['roll'] = {"long_name": "roll", "units": "degrees"}
            elif "EGIM" in key and "SD_capacity" in key:
                # Obtain the Slot number
                slot_number = key.split("_")[1].split("t")[1]
                wf.data.rename(columns={key: "SD{}".format(slot_number)},
                               inplace=True)
                wf.data["SD{}_QC".format(slot_number)] = 0
                wf.meaning["SD{}".format(slot_number)] = {
                    "long_name": "slot_SD_capacity", "units": "kBytes"}
            elif "EGIM" in key and "_current" in key:
                # Obtain the Slot number
                slot_number = key.split("_")[1].split("t")[1]
                wf.data.rename(columns={key: "current{}".format(slot_number)},
                               inplace=True)
                wf.data["current{}_QC".format(slot_number)] = 0
                wf.meaning["current{}".format(slot_number)] = {
                    "long_name": "slot_current", "units": "mA"}
            elif key == "waterInstrusion":
                wf.data.rename(columns={"waterInstrusion": "leak"},
                               inplace=True)
                wf.data["leak_QC"] = 0
                wf.meaning['leak'] = {"long_name": "WaterIntrusion",
                                      "units": "bool"}
            elif "EGIM" in key and "_temperature" in key:
                # Obtain the Slot number
                slot_number = key.split("_")[1].split("t")[1]
                wf.data.rename(columns={key: "temperature{}".format(
                    slot_number)}, inplace=True)
                wf.data["temperature{}_QC".format(slot_number)] = 0
                wf.meaning["temperature{}".format(slot_number)] = {
                    "long_name": "EGIM_slot_temperature",
                    "units": "degree Celsius"}
            elif "EGIM" in key and "_pressure" in key:
                # Obtain the Slot number
                slot_number = key.split("_")[1].split("t")[1]
                wf.data.rename(columns={key: "pressure{}".format(slot_number)},
                               inplace=True)
                wf.data["pressure{}_QC".format(slot_number)] = 0
                wf.meaning["pressure{}".format(slot_number)] = {
                    "long_name": "EGIM_slot_pressure", "units": "mBars"}
            elif key == "voltage":
                wf.data.rename(columns={"voltage": "voltage"}, inplace=True)
                wf.data["voltage_QC"] = 0
                wf.meaning['voltage'] = {"long_name": "Incoming EGIM voltage",
                                         "units": "volts"}
            elif key == "energy":
                wf.data.rename(columns={"energy": "energy"}, inplace=True)
                wf.data["energy_QC"] = 0
                wf.meaning['energy'] = {"long_name": "Energy consumption",
                                        "units": "mAh"}
            elif key == "Comment":
                del wf.data["Comment"]
            elif key == "Temperature [C]":
                wf.data.rename(columns={"Temperature [C]": "TEMP"},
                               inplace=True)
                wf.data["TEMP_QC"] = 0
                wf.meaning['TEMP'] = {"long_name": "sea_water_temperature",
                                      "units": "degree Celsius"}
            elif key == "Humidity [%]":
                wf.data.rename(columns={"Humidity [%]": "RELH"}, inplace=True)
                wf.data["RELH_QC"] = 0
                wf.meaning['RELH'] = {"long_name": "RELH", "units": "%"}
            elif key == "Sequence #":
                wf.data.rename(columns={"Sequence #": "Sequence"},
                               inplace=True)
                wf.meaning['Sequence'] = {"long_name": "Sequence number",
                                          "units": "#"}
            elif key == "Data Points":
                wf.data.rename(columns={"Data Points": "Data Points"},
                               inplace=True)
                wf.meaning['Data Points'] = {"long_name": "Data Points",
                                             "units": "#"}

        wf.metadata = metadata

        return wf

    @staticmethod
    def to_netcdf(observatory, instrument, data, path, qc_tests=True,
                  only_qc1=False):
        """It creates a netCDF3 file following the OceanSites standard.

        Parameters
        ----------
            observatory: str
                EGIM observatory name.
            instrument: str
                Instrument name.
            data: Pandas dataframe, WaterFrame
                Data to be saved into a netCDF file.
            path: str
                Path to save the netCDF file.
            qc_tests: bool (optional, qc_tests=True)
                It indicates if QC test should be passed.
            only_qc1: bool (optional, only_qc1=True)
                It indicates to save only values with QC = 1.

        Returns
        -------
            True: bool
                Operation successful.

        """
        # Choose observatory metadata
        if observatory == 'EMSODEV-EGIM-node00001':
            metadata = EGIM.METADATA_00001

        # Creation of WaterFrame
        if type(data).__name__ == 'WaterFrame':
            wf = data
        else:
            wf = EGIM.to_waterframe(data, metadata)

        if qc_tests:
            # Creation of QC Flags following OceanSites recomendation
            for parameter in wf.parameters():
                # Reset QC Flags to 0
                wf.reset_flag(parameters=parameter, flag=0)
                # Flat test
                wf.flat_test(parameters=parameter, window=0, flag=4)
                # Spike test
                wf.spike_test(parameters=parameter, window=0, threshold=3, flag=4)
                # Range test
                wf.range_test(parameters=parameter, flag=4)
                # Change flags from 0 to 1
                wf.flag2flag(parameters=parameter, original_flag=0, translated_flag=1)
        if only_qc1:
            wf.use_only(flags=1)

        # Creation of an xarray dataset
        ds = xr.Dataset(data_vars=wf.data, attrs=metadata)

        # Creation of attr of dataset parameters
        for key in wf.data.keys():
            if key == 'TEMP':
                if observatory == 'EMSODEV-EGIM-node00001':
                    if instrument == '37-14998':
                        attr = EGIM.METADATA_TEMP_SBE37
                    elif instrument == '4381-606':
                        attr = EGIM.METADATA_TEMP_AADI4381
            if key == 'PSAL':
                if observatory == 'EMSODEV-EGIM-node00001':
                    attr = EGIM.METADATA_PSAL_SBE37
            if key == 'SVEL':
                if observatory == 'EMSODEV-EGIM-node00001':
                    attr = EGIM.METADATA_SVEL_SBE37
            if key == 'CNDC':
                if observatory == 'EMSODEV-EGIM-node00001':
                    attr = EGIM.METADATA_CNDC_SBE37
            if key == 'MPMN':
                if observatory == 'EMSODEV-EGIM-node00001':
                    attr = EGIM.METADATA_MPMN_SBE37
            if key == 'PRES':
                if observatory == 'EMSODEV-EGIM-node00001':
                    attr = EGIM.METADATA_PRES_SBE54
            if key == 'TUR4':
                if observatory == 'EMSODEV-EGIM-node00001':
                    attr = EGIM.METADATA_TUR4_NTURTD
            if key == 'OSAT':
                if observatory == 'EMSODEV-EGIM-node00001':
                    attr = EGIM.METADATA_OSAT_AADI4381
            if key == 'DOX2':
                if observatory == 'EMSODEV-EGIM-node00001':
                    attr = EGIM.METADATA_DOX2_AADI4381
            if key == 'TEMP_QC':
                attr = EGIM.METADATA_TEMP_QC
            if key == 'PSAL_QC':
                attr = EGIM.METADATA_PSAL_QC
            if key == 'SVEL_QC':
                attr = EGIM.METADATA_SVEL_QC
            if key == 'CNDC_QC':
                attr = EGIM.METADATA_CNDC_QC
            if key == 'MPMN_QC':
                attr = EGIM.METADATA_MPMN_QC
            if key == 'PRES_QC':
                attr = EGIM.METADATA_PRES_QC
            if key == 'TUR4_QC':
                attr = EGIM.METADATA_TUR4_QC
            if key == 'OSAT_QC':
                attr = EGIM.METADATA_OSAT_QC
            if key == 'DOX2_QC':
                attr = EGIM.METADATA_DOX2_QC
            # Add the attrs to the variable
            ds[key].attrs = attr

        # Creation of the nc file
        ds.to_netcdf(path, format="NETCDF3_64BIT")

        return True

    @staticmethod
    def to_csv(observatory, data, path, qc_tests=True, only_qc1=False):
        """It creates a CSV file following the OceanSites standard.

        Parameters
        ----------
            observatory: str
                EGIM observatory name.
            instrument: str
                Instrument name.
            data: Pandas dataframe, WaterFrame
                Data to be saved into a netCDF file.
            path: str
                Path to save the CSV file.
            qc_tests: bool (optional, qc_tests=True)
                It indicates if QC test should be passed.
            only_qc1: bool (optional, only_qc1=True)
                It indicates to save only values with QC = 1

        Returns
        -------
            True: bool
                Operation successful.
        """
        # Choose observatory metadata
        if observatory == 'EMSODEV-EGIM-node00001':
            metadata = EGIM.METADATA_00001

        # Creation of WaterFrame
        if type(data).__name__ == 'WaterFrame':
            wf = data
        else:
            wf = EGIM.to_waterframe(data, metadata)

        if qc_tests:
            # Creation of QC Flags following OceanSites recomendation
            for parameter in wf.parameters():
                # Reset QC Flags to 0
                wf.reset_flag(parameters=parameter, flag=0)
                # Flat test
                wf.flat_test(parameters=parameter, window=0, flag=4)
                # Spike test
                wf.spike_test(parameters=parameter, window=0, threshold=3, flag=4)
                # Range test
                wf.range_test(parameters=parameter, flag=4)
                # Change flags from 0 to 1
                wf.flag2flag(parameters=parameter, original_flag=0, translated_flag=1)
        if only_qc1:
            wf.use_only(flags=1)

        # Metadata text creation
        metadata_text = "# Global attributes;Value"
        for key, value in metadata.items():
            metadata_text += "# " + str(key) + ";" + str(value) + "\n"

        # Creation of the csv file of the data
        wf.data.to_csv(path, sep=";", header=True)

        # Adding the metadata to the file
        with open(path, 'r+') as f:
            content = f.read()
            f.seek(0, 0)
            f.write(metadata_text.rstrip('\r\n') + '\n\n\n\n' + content)

        return True

    @staticmethod
    def from_raw_csv(observatory, path, qc_tests=False):
        """ It opens a csv file that contains data from an EGIM instrument.

        Parameters
        ----------
            observatory: str
                Name of the observatory. For now, only "EMSO-Azores" is
                possible.
                Write "EMSO-Azores" if the node is "EMSO-Azores" and source
                files are from: www.seanoe.org
            path: str
                Path where the file is.
            qc_tests: bool (optional)
                It indicates if QC test should be passed.

        Returns
        -------
            wf: WaterFrame
        """
        # Creation of a WaterFrame
        wf = WaterFrame()

        # The arguments of pandas.read_csv could change depending on the
        # source.
        sep = ","  # Delimiter to use.
        header = None  # Row number(s) to use as the column names, and the
        # start of the data.
        skiprows = None  # Line numbers to skip (0-indexed) or number of
        # lines to skip (int) at the start of the file.
        if observatory == "EMSO-Azores":
            sep = ";"
            header = 1
            skiprows = [2]
            format_time = '%d/%m/%Y %H:%M:%S'

            # Add metadata info
            wf.metadata = EGIM.METADATA_AZORES.copy()
        else:
            warnings.warn("EGIM cannot open the CSV because it does not know the observatory.")
            return
        # Load data from CSV into a DataFrame
        df = pd.read_csv(path, sep=sep, header=header, skiprows=skiprows)
        columns = df.columns.tolist()  # get the columns
        cols_to_use = columns[:len(columns)-1]  # drop the last one
        df = pd.read_csv(path, sep=sep, header=header, skiprows=skiprows,
                         usecols=cols_to_use)

        # Create the time index
        df['TIME'] = pd.to_datetime(df['DATE'] + ' ' + df['TIME'],
                                    format=format_time)
        df.set_index(pd.DatetimeIndex(df['TIME']), inplace=True)
        df.drop(['DATE', 'TIME'], inplace=True, axis=1)

        # Add QC columns
        for parameter in df.keys():
            df["{}_QC".format(parameter)] = 0

        # Add DataFrame into the WaterFrame
        wf.data = df.copy()

        # Change column names and add meanings
        for parameter in wf.parameters():
            if parameter == "TEMPERATURE":
                info = {'long_name': EGIM.METADATA_TEMP_SBE37['long_name'],
                        'units': EGIM.METADATA_TEMP_SBE37['units']}
                wf.meaning[parameter] = info.copy()
                wf.rename("TEMPERATURE", "TEMP")
            elif parameter == "CONDUCTIVITY":
                info = {'long_name': EGIM.METADATA_CNDC_SBE37['long_name'],
                        'units': EGIM.METADATA_CNDC_SBE37['units']}
                wf.meaning[parameter] = info.copy()
                wf.rename("CONDUCTIVITY", "CNDC")
            elif parameter == "PRESSION":
                info = {'long_name': EGIM.METADATA_MPMN_SBE37['long_name'],
                        'units': EGIM.METADATA_MPMN_SBE37['units']}
                wf.meaning[parameter] = info.copy()
                wf.rename("PRESSION", "MPMN")
            elif parameter == "SALINITY":
                info = {'long_name': EGIM.METADATA_PSAL_SBE37['long_name'],
                        'units': EGIM.METADATA_PSAL_SBE37['units']}
                wf.meaning[parameter] = info.copy()
                wf.rename("SALINITY", "PSAL")
            elif parameter == "SOUND SPEED":
                info = {'long_name': EGIM.METADATA_SVEL_SBE37['long_name'],
                        'units': EGIM.METADATA_SVEL_SBE37['units']}
                wf.meaning[parameter] = info.copy()
                wf.rename("SOUND SPEED", "SVEL")
            elif parameter == "OXYGEN":
                info = {'long_name': EGIM.METADATA_DOX2_AADI4381['long_name'],
                        'units': EGIM.METADATA_DOX2_AADI4381['units']}
                wf.meaning[parameter] = info.copy()
                wf.rename("OXYGEN", "DOX2")
            elif parameter == "SATURATION":
                info = {'long_name': EGIM.METADATA_OSAT_AADI4381['long_name'],
                        'units': EGIM.METADATA_OSAT_AADI4381['units']}
                wf.meaning[parameter] = info.copy()
                wf.rename("SATURATION", "OSAT")
            elif parameter == "NTU_SIGNAL_2":
                info = {'long_name': EGIM.METADATA_TUR4_NTURTD['long_name'],
                        'units': EGIM.METADATA_TUR4_NTURTD['units']}
                wf.meaning[parameter] = info.copy()
                wf.rename("NTU_SIGNAL_2", "TUR4")
            elif parameter == "PRESSURE":
                info = {'long_name': EGIM.METADATA_PRES_SBE54['long_name'],
                        'units': EGIM.METADATA_PRES_SBE54['units']}
                wf.meaning[parameter] = info.copy()
                wf.rename("PRESSURE", "PRES")
            else:
                # Delete the parameter
                wf.drop(parameter)

        # Creation of QC Flags following OceanSites recomendation
        if qc_tests:
            # Reset QC Flags to 0
            wf.reset_flag(flag=0)
            # Flat test
            wf.flat_test(window=0, flag=4)
            # Spike test
            wf.spike_test(window=0, threshold=3, flag=4)
            # Range test
            wf.range_test(flag=4)
            # Change flags from 0 to 1
            wf.flag2flag(original_flag=0, translated_flag=1)

        return wf



"""
Module to open data from the TOB files of Sea And Sun CTD
"""
import pandas as pd
from mooda import WaterFrame


class SeaAndSun:
    """
    Class to import TOB files from Sea And Sun.
    """

    @staticmethod
    def from_tob_to_waterframe(path, qc_tests=False):
        """
        It reads a TOB file and parses its content into a WaterFrame.

        Parameters
        ----------
            path: str
                Path where the TOB files is.
            qc_tests: bool (optional, qc_tests=False)
                If true, it executes WaterFrame.qc().

        Returns
        -------
            wf: WaterFrame
        """
        # Find the number of line where starts data and column names
        # Read file
        lines = ""
        with open(path) as handle:
            lines = handle.readlines()

        data_start = None
        column_names = []
        for index, line in enumerate(lines):
            line_strip = line.strip()
            if line_strip.startswith('; Datasets'):
                # Column line
                line_split = line_strip.split(' ')
                column_names = [
                    column for column in line_split if column not in [';', '', 'Datasets']]
                data_start = index + 3
            elif line_strip.startswith('1 '):
                data_start = index
                break
        # Change name of columns and find date and time positions
        parameters = ['TIME']
        date_column = None
        time_column = None
        for index, name in enumerate(column_names):
            if name == 'Vbatt':
                parameters.append('BATT')
            elif name == 'Press':
                parameters.append('PRES')
            elif name == 'Temp':
                parameters.append('TEMP')
            elif name == 'SALIN':
                parameters.append('PSAL')
            elif name == 'Cond':
                parameters.append('CNDC')
            elif name == 'IntD':
                date_column = index + 1
            elif name == 'IntT':
                time_column = index + 1

        # Create dataframe
        df = pd.read_csv(path, sep=" ", skiprows=range(data_start), skipinitialspace=True,
                         header=None, parse_dates=[[date_column, time_column]])
        del df[0]
        df.columns = parameters
        df.set_index('TIME', inplace=True)

        # Creation of wf
        wf = WaterFrame(df=df)

        # Adding meaning
        for parameter in wf.parameters():
            if parameter == 'BATT':
                wf.meaning['BATT'] = {
                    'long_name': 'Remaining battery',
                    'units': 'Volt'}
            elif parameter == 'PRES':
                wf.meaning['PRES'] = {
                    'long_name': 'Sea water pressure',
                    'units': 'dBar'}
            elif parameter == 'TEMP':
                wf.meaning['TEMP'] = {
                    'long_name': 'Sea water temperature',
                    'units': 'degree Celsius'}
            elif parameter == 'TEMP':
                wf.meaning['CNDC'] = {
                    'long_name': 'Sea water conductivity',
                    'units': 'mS/cm'}
            elif parameter == 'TEMP':
                wf.meaning['PSAL'] = {
                    'long_name': 'Partial Salinity',
                    'units': 'PSU'}

        if qc_tests:
            wf.qc()

        return wf



"""
Module to open data from the cnv files from Sea-Bird CTD
"""
from datetime import datetime
import pandas as pd
from mooda import WaterFrame


class SeaBird:
    """
    Class to import cnv files from Sea-Bird.
    """

    @staticmethod
    def from_cnv_to_waterframe(path, qc_tests=False):
        """
        It reads a cnv file and parses its content into a WaterFrame.

        Parameters
        ----------
            path: str
                Path where the TOB files is.
            qc_tests: bool (optional, qc_tests=False)
                If true, it executes WaterFrame.qc().

        Returns
        -------
            wf: WaterFrame
        """

        # Read file
        lines = ""
        with open(path) as handle:
            lines = handle.readlines()

        # Extract info from cnv
        metadata = {}
        meaning = {}
        data_start = None
        column_names = []
        for index, line in enumerate(lines):
            line_strip = line.strip()
            if line_strip.startswith('* Sea-Bird '):
                # Metadata model line
                # ex: "* Sea-Bird SBE25 Data File:"
                line_split = line_strip.split(' ')
                metadata['model'] = line_split[2]
            elif line_strip.startswith('* Software Version '):
                # Sw version line
                # ex: "* Software Version 1.59"
                line_split = line_strip.split(' ')
                metadata['software_version'] = line_split[3]
            elif line_strip.startswith('* Temperature SN = '):
                # ex: "* Temperature SN = 031485"
                line_split = line_strip.split(' ')
                metadata['temperature_sn'] = line_split[4]
            elif line_strip.startswith('* Conductivity SN = '):
                # ex: "* Conductivity SN = 041189"
                line_split = line_strip.split(' ')
                metadata['conductivity_sn'] = line_split[4]
            elif line_strip.startswith('* System UpLoad Time = '):
                # ex: "* System UpLoad Time = Mar 18 2019 10:37:17"
                line_split = line_strip.split('* System UpLoad Time = ')
                metadata['system_upLoad_time'] = line_split[1]
            elif line_strip.startswith('* battery type = '):
                # ex: "* battery type = ALKALINE"
                line_split = line_strip.split(' ')
                try:
                    metadata['battery_type'] = line_split[4]
                except IndexError:
                    pass
            elif line_strip.startswith('# name '):
                # ex: "# name 0 = scan: Scan Count"
                line_split = line_strip.split(' ')
                name = line_split[4][:-1]  # Remove last ":"
                if name == 'prSM':
                    column_names.append('PRES')
                    meaning['PRES'] = {
                        'long_name': 'Strain Gauge Pressure',
                        'units': 'db'
                    }
                elif name == 't090C':
                    column_names.append('TEMP')
                    meaning['TEMP'] = {
                        'long_name': 'Sea Water Temperature',
                        'units': 'deg C'
                    }
                elif name == 'sal00':
                    column_names.append('PSAL')
                    meaning['PSAL'] = {
                        'long_name': 'Partial Salinity',
                        'units': 'PSU'
                    }
                elif name == 'seaTurbMtr':
                    column_names.append('TUR4')
                    meaning['PSAL'] = {
                        'long_name': 'Turbidity',
                        'units': 'FTU'
                    }
                elif name == 'wetStar':
                    column_names.append('FLUO')
                    meaning['FLUO'] = {
                        'long_name': 'Fluorescence',
                        'units': 'mg/m^3'
                    }
                elif name == 'c0mS/cm':
                    column_names.append('CNDC')
                    meaning['CNDC'] = {
                        'long_name': 'Conductivity',
                        'units': 'mS/cm'
                    }
                else:
                    column_names.append(name)
            elif line_strip.startswith('# start_time = '):
                # ex: "# name 0 = scan: Scan Count"
                line_split = line_strip.split('# start_time = ')
                metadata['start_time'] = line_split[1]
            elif line_strip.startswith('*END*'):
                data_start = index + 1

        # Create dataframe
        df = pd.read_csv(path, sep=" ", skiprows=range(data_start), skipinitialspace=True,
                         header=None)
        df.columns = column_names

        # Creation of TIME index
        start = pd.to_datetime(metadata['start_time']).to_pydatetime()
        seconds = (start - datetime(1970, 1, 1)).total_seconds()
        df['TIME'] = pd.to_datetime(df['timeS'] + seconds, unit='s')
        df.set_index('TIME', inplace=True)

        # Delete non used columns
        del df['timeS']
        del df['scan']
        del df['flag']

        # Creation of wf
        wf = WaterFrame(df=df, metadata=metadata, meaning=meaning)

        if qc_tests:
            wf.qc()

        return wf



""" Module to access to Pangea (https://www.pangaea.de/)
Adaptation of code from https://github.com/huberrob/panpython by Robert Huber and Markus Stocker
"""
import warnings
from xml.etree import ElementTree
import io
import re
import requests
import pandas as pd
from mooda import WaterFrame


class Pangaea:
    """ Main class of the module"""

    def __init__(self, id_data=None):
        """
        Constructor

        Parameters
        ----------
            id_data: int
                Id of the datasource.
        """
        self.id = 0  # pylint: disable=C0103
        self.data = None

        if id_data is None:
            self.metadata = dict()
            self.data = pd.DataFrame()
        else:
            if isinstance(id_data, int):
                self.id = id_data
            else:
                raise TypeError('id_data must be an integer.')

            # Add metadata to self.metadata
            self.metadata = self.get_metadata(self.id)

            # Add data to self.data
            self.data = self.get_data(id_data)

            print(self.data.head())

    @staticmethod
    def get_metadata(id_data):
        """
        It calls to the Pangea API to obtain metadata information.

        Parameters
        ----------
            id_data: int
                Id of the data that you want to obtain the metadata.

        Returns
        -------
            metadata: dict
                Dictionary with metadata information

        """
        metadata_url = f"https://doi.pangaea.de/10.1594/PANGAEA.{id_data}?format=metainfo_xml"
        metadata = dict()

        response = requests.get(metadata_url)
        if response.status_code != 404:
            xml = ElementTree.fromstring(response.text)
            namespace = {'md': 'http://www.pangaea.de/MetaData'}

            # Read metadata
            metadata['login_status'] = xml.find('./md:technicalInfo/md:entry[@key="loginOption"]',
                                                namespace).get('value')
            if metadata['login_status'] != 'unrestricted':
                warnings.warn(f"Dataset with id {id_data} is protected")
            try:
                metadata['hierarchy_level'] = xml.find(
                    './md:technicalInfo/md:entry[@key="hierarchyLevel"]', namespace).get('value')
                if metadata['hierarchy_level'] == "parent":
                    warnings.warn(f"Dataset with id {id_data} is of type parent")
            except AttributeError:
                metadata['hierarchy_level'] = "child"
            metadata['title'] = xml.find("./md:citation/md:title", namespace).text
            metadata['year'] = xml.find("./md:citation/md:year", namespace).text
            metadata['doi'] = xml.find("./md:citation/md:URI", namespace).text
            metadata['topotype'] = xml.find("./md:extent/md:topoType", namespace).text
            metadata['authors'] = [", ".join((author.find("md:lastName", namespace).text,
                                              author.find("md:firstName", namespace).text))
                                   for author in xml.findall("./md:citation/md:author", namespace)]
            # Parameters
            metadata['parameters'] = dict()
            short_names = dict()
            xml_matrix_column = xml.findall("./md:matrixColumn", namespace)
            for matrix in xml_matrix_column:
                xml_type = matrix.get('type')
                source = matrix.get('source')
                xml_parameter = matrix.find("md:parameter", namespace)
                # Name
                name = xml_parameter.find('md:name', namespace).text
                # Short name
                short_name = xml_parameter.find('md:shortName', namespace).text
                # Rename duplicate short_name
                if short_name in short_names:
                    short_name += f"_{short_names[short_name]}"
                    short_names[short_name] += 1
                else:
                    short_names[short_name] = 1
                # Unit
                try:
                    unit = xml_parameter.find('md:unit', namespace).text
                except AttributeError:
                    if name == "DATE/TIME":
                        pass
                    else:
                        warnings.warn(f"Parameter {name} without units")
                    unit = None
                # Parameter id
                try:
                    id_parameter = xml_parameter.get('id').split(".")[2]
                except AttributeError:
                    warnings.warn(f'Parameter "{name}" without id')
                    id_parameter = None

                parameter_dict = {'type': xml_type,
                                  'source': source,
                                  'name': name,
                                  'short_name': short_name,
                                  'units': unit,
                                  'id': id_parameter}
                metadata['parameters'][short_name] = parameter_dict

            # Events
            xml_events = xml.findall("./md:event", namespace)
            metadata['events'] = [{'label': event.find('./md:label', namespace).text,
                                   'latitude': event.find('./md:latitude', namespace).text,
                                   'longitude': event.find('./md:longitude', namespace).text,
                                   'elevation': event.find('./md:elevation', namespace).text,
                                   'date_time': event.find('./md:dateTime', namespace).text
                                   } for event in xml_events]

        else:
            warnings.warn(f"Dataset with id {id_data} does not exist")

        return metadata

    @staticmethod
    def get_data(id_data):
        """
        Obtain the data from Pangea data with the input id.

        Parameters
        ----------
            id_data: int
                Id of the data source.

        Returns
        -------
            df: pandas.DataFrame
        """
        data_url = f"https://doi.pangaea.de/10.1594/PANGAEA.{id_data}?format=textfile"
        xml = requests.get(data_url).text
        data = re.sub(r"/\*(.*)\*/", "", xml, 1, re.DOTALL).strip()
        df_pangea = pd.read_csv(io.StringIO(data), index_col=False, error_bad_lines=False,
                                sep=u'\t')

        # Replace Quality Flags
        df_pangea.replace(regex=r'^[\?/\*#\<\>]', value='', inplace=True)
        # Adjust Column Data Types
        df_pangea = df_pangea.apply(pd.to_numeric, errors='ignore')
        if 'Date/Time' in df_pangea.columns:
            df_pangea['Date/Time'] = pd.to_datetime(df_pangea['Date/Time'],
                                                    format='%Y/%m/%dT%H:%M:%S')

        return df_pangea

    def to_waterframe(self, id_data=None, data=None, metadata=None):
        """
        It creates a mooda.WaterFrame object.

        Parameters
        ----------
            id_data: int (optional, id_data=None)
                Id if the dataset of pangea to use.
            data: pandas.DataFrame (optional, data=None)
                DataFrame to include into the WaterFrame.
            metadata: dict (optional, metadata=None)
                Metadata to include into the WaterFrame.

        Returns
        -------
            wf_pangea: mooda.WaterFrame
                WaterFrame object.
        """
        # Check where to find data and metadata
        if id_data:
            metadata_pangea = self.get_metadata(id_data)
            data_pangea = self.get_data(id_data)
        else:
            if data:
                data_pangea = data
            else:
                data_pangea = self.data
            if metadata:
                metadata_pangea = metadata
            else:
                metadata_pangea = self.metadata

        # Change names of columns
        # Delete the units from the column names (ex: Elevation [m] -> Elevation)
        # and Date/Time to TIME
        # Delete "Event" column
        for key in data_pangea.keys():
            if "[" in key:
                data_pangea.rename(columns={key: key.split(" [")[0]}, inplace=True)
            elif key == "Date/Time":
                data_pangea.rename(columns={key: "TIME"}, inplace=True)
            elif key == "Event":  # TODO: Check what is "Event"
                del data_pangea["Event"]

        # Set index to TIME
        try:
            data_pangea.set_index("TIME", inplace=True)
        except KeyError:
            warnings.warn('There is no "TIME" in the columns of the dataframe. The WaterFrame' +
                          ' will not contain a time index.')

        # Set meaning
        meaning_pangea = metadata_pangea['parameters']
        # Add "long_name" to the dictionary if it does not exist
        for meaning in meaning_pangea:
            keys = meaning_pangea[meaning].keys()
            if "long_name" not in keys and "name" in keys:
                meaning_pangea[meaning]["long_name"] = meaning_pangea[meaning]["name"]

        wf_pangea = WaterFrame(df=data_pangea, metadata=metadata_pangea, meaning=meaning_pangea)

        return wf_pangea

    @staticmethod
    def from_source_to_waterframe(id_data=None, data=None, metadata=None):
        """
        It creates a mooda.WaterFrame object.

        Parameters
        ----------
            id_data: int (optional, id_data=None)
                Id if the dataset of pangea to use.
            data: pandas.DataFrame (optional, data=None)
                DataFrame to include into the WaterFrame.
            metadata: dict (optional, metadata=None)
                Metadata to include into the WaterFrame.

        Returns
        -------
            wf_pangea: mooda.WaterFrame
                WaterFrame object.
        """
        # Check where to find data and metadata
        if id_data:
            metadata_pangea = Pangaea.get_metadata(id_data)
            data_pangea = Pangaea.get_data(id_data)
        else:
            if data and metadata:
                data_pangea = data
                metadata_pangea = metadata
            else:
                warnings.warn("You must insert id_data or data and metadata.")
                return False

        # Change names of columns
        # Delete the units from the column names (ex: Elevation [m] -> Elevation)
        # and Date/Time to TIME
        # Delete "Event" column
        for key in data_pangea.keys():
            if "[" in key:
                data_pangea.rename(columns={key: key.split(" [")[0]}, inplace=True)
            elif key == "Date/Time":
                data_pangea.rename(columns={key: "TIME"}, inplace=True)
            elif key == "Event":  # TODO: Check what is "Event"
                del data_pangea["Event"]

        # Set index to TIME
        try:
            data_pangea.set_index("TIME", inplace=True)
        except KeyError:
            warnings.warn('There is no "TIME" in the columns of the dataframe. The WaterFrame' +
                          ' will not contain a time index.')

        # Set meaning
        meaning_pangea = metadata_pangea['parameters'].copy()
        # Add "long_name" to the dictionary if it does not exist
        for meaning in meaning_pangea:
            keys = meaning_pangea[meaning].keys()
            if "long_name" not in keys and "name" in keys:
                meaning_pangea[meaning]["long_name"] = meaning_pangea[meaning]["name"]
        # Delete 'parameters' and 'events' from the metadata
        del metadata_pangea['parameters']
        del metadata_pangea['events']

        wf_pangea = WaterFrame(df=data_pangea, metadata=metadata_pangea, meaning=meaning_pangea)

        return wf_pangea



import matplotlib.pyplot as plt
from mpl_toolkits.basemap import Basemap


class PlotMap:
    """
    It contains functions related to the management of maps.
    It creates an instance variable called 'm' that is a Basemap object.
    We are going to use and modify 'm' in all functions.
    """
    def __init__(self, ):
        self.m = Basemap()

    def map_world(self, res='l'):
        """
        It creates a map of the world and saves into 'm'.

        Parameters
        ----------
            res: str, 'l', 'i', 'h', 'f', optional (res = 'l')
                Resolution of boundary database to use. Can be c (crude),
                l (low), i (intermediate), h (high), f (full) or None. If None,
                no boundary data will be read in (and class methods such as
                draw coastlines will raise an if invoked). Higher res datasets
                are much slower to draw.
        """
        self.m = Basemap(projection='mill', resolution=res)
        # Draw the coastal lines
        self.m.drawcoastlines()
        # Fill the continents with black
        self.m.fillcontinents(color='K')

    def map_mediterranean(self, res='l'):
        """
        It creates a map of the Mediterranean.

        Parameters
        ----------
            res: str, 'l', 'i', 'h', 'f', optional (res = 'l')
                Resolution of boundary database to use. Can be c (crude),
                l (low), i (intermediate), h (high), f (full) or None. If None,
                no boundary data will be read in (and class methods such as
                drawcoastlines will raise an if invoked). Higher res datasets
                are much slower to draw.
        """
        self.m = Basemap(projection='mill', resolution=res, llcrnrlat=30,
                         llcrnrlon=-13, urcrnrlat=47, urcrnrlon=38)
        # Draw the coastal lines
        self.m.drawcoastlines()
        # Fill the continents with black
        self.m.fillcontinents(color='K')

    def add_point(self, lon, lat, dot_color="blue", label=None,
                  label_color="Green"):
        """
        It adds points to the map.

        Parameters
        ----------
            lon: float
                Longitude.
            lat: float
                Latitude.
            dot_color: str
                Color of the point.
            label: str
                Text to write in the point.
            label_color: str
                Color of the label.
        """
        x, y = self.m(float(lon), float(lat))
        self.m.plot(x, y, color=dot_color, marker='o', markersize=7)
        if label is not None:
            plt.text(x+10000, y+5000, label, color=label_color)



"""Module to read data from a Hobo instrument file.
Initial work from Carlos Rodero Garcia ()
"""
from io import StringIO
import csv
import re
import pandas as pd
from mooda import WaterFrame


class Hobo:
    """Class to import Hobo data from a csv file"""

    SN_REGEX = re.compile(r'(?:LGR S/N: |Serial Number:)(\d+)')

    @staticmethod
    def _find_units(header):
        """
        Internal function to find the units value from the header of a HOBO csv file.

        Parameters
        ----------
            header: str
                Header of one parameter.

        Returns
        -------
            units: str
                Units of the parameter.
        """
        header_units = header.split(",", 2)[1].strip()
        units = header_units.split(" ")[0].strip()
        return units

    @staticmethod
    def _find_name(header):
        """
        Infernal function. It looks for the label of the parameter in the header.

        Parameters
        ----------
            header: str
                Header of one parameter.

        Returns
        -------
            name: str
                Label of the parameter.
        """
        header_name = header.rsplit(",", 1)[1].strip(" )")
        if header_name.split(":")[0].strip() == "LBL":
            name = header_name.split(":")[1].strip()
        else:
            name = header.split(",", 1)[0].strip()
        return name

    @staticmethod
    def _find_long_name(header):
        """
        Infernal function. It looks for the long name of the parameter in the header.

        Parameters
        ----------
            header: str
                Header of one parameter.

        Returns
        -------
            long_name: str
                Long name of the parameter.
        """
        long_name = header.split(",", 1)[0].strip()
        return long_name

    @staticmethod
    def from_csv_to_waterframe(path, qc_tests=False):
        """
        It creates a WaterFrame form data of the csv file of a hobo logger.

        Parameters
        ----------
            path: str
                Path where the csv is located.
            qc_tests: bool (optional)
                Make the QC test of the WaterFrame before return it.

        Returns
        -------
            wf: WaterFrame
                WaterFrame with the data and metadata of the source csv file.
        """

        wf = WaterFrame()

        with open(path, encoding='utf-8-sig') as f_hobo:

            # Find title
            title = next(f_hobo)
            # Get title, we supposes that the title is in the first line and it is something like
            # ['"Título de trazado', ' test_lab"']
            wf.metadata['title'] = title.strip().split(":")[1].strip(" \"")

            # Find headers
            header = next(f_hobo)
            # Get serial number
            sn_match = Hobo.SN_REGEX.search(header)
            if sn_match:
                wf.metadata['sn'] = sn_match.groups()[0]

            # Find and set column names for headers
            headers = next(csv.reader(StringIO(header)))
            timestamp_position = None
            for i, header in enumerate(headers):
                # Find the column of the timestamp
                if 'Date Time' in header or 'Fecha Tiempo' in header:
                    timestamp_position = i
                    continue

                # Find the temperature
                temp_vocabulary = ['High Res. Temp.', 'High-Res Temp', 'Temp,', 'Temp.',
                                   'Temperature']
                if any(word in header for word in temp_vocabulary):
                    wf.meaning[header] = {
                        'long_name': Hobo._find_long_name(header),
                        'units': Hobo._find_units(header),
                        'name': 'temp'
                    }
                    continue

                # Find pressure
                if 'Pres abs,' in header:
                    wf.meaning[header] = {
                        'long_name': Hobo._find_long_name(header),
                        'units': Hobo._find_units(header),
                        'name': 'press'
                    }
                    continue

                # Find rh
                if 'RH,' in header:
                    wf.meaning[header] = {
                        'long_name': Hobo._find_long_name(header),
                        'units': Hobo._find_units(header),
                        'name': 'rh'
                    }
                    continue

                # Find battery
                if 'Batt, V' in header:
                    wf.meaning[header] = {
                        'long_name': Hobo._find_long_name(header),
                        'units': Hobo._find_units(header),
                        'name': 'batt'
                    }
                    continue

            if timestamp_position:
                wf.from_csv(path=f_hobo, names=headers, index_col=timestamp_position)

                # Rename columns
                for parameter in wf.parameters():
                    renamed = False
                    for key, value in wf.meaning.items():
                        if key == parameter:
                            wf.rename(parameter, value['name'])
                            renamed = True
                            break
                    if not renamed:
                        wf.drop(parameter)
                # Rename index to TIME
                wf.data.set_index(pd.DatetimeIndex(wf.data.index), inplace=True)
                wf.data.index.names = ['TIME']

                if qc_tests:
                    wf.qc()
            else:
                raise ValueError("Not possible to read the structure of the csv file.")
        return wf



"""It just contains the WaterFrame class"""
# pylint: disable=C0103, too-many-lines
import copy
import io
import pickle
import sys
import datetime
import warnings
import json
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
import pandas as pd


class WaterFrame:
    """Object to manage data series from marine observatories.
    Its most important instance variables are the following:

    data: A pandas DataFrame that contains the measurement values of the
    timeserie.

    metadata: A dictionary that contains the metadata information of the
    timeserie."""

    def __init__(self, path=None, df=None, metadata=None, meaning=None):
        """It creates the instance following variables:
        data -- A pandas DataFrame that contains the measurement values of the
        time series.
        metadata -- A dictionary that contains the metadata information of the
        time series.
        meaning -- A dictionary that contains the meaning of the keys of data
        (i.e. "TEMP": "Sea water temperature")

        Parameters
        ----------
            path: str, optional
                Create a WaterFrame with the data of the file of the path.
                The file must be a NetCDF, a CSV or a JSON string
            df: pandas Dataframe
                pandas DataFrame
            metadata: dict
                Metadata dictionary
            meaning: dict
                Meaning dictionary
        """
        # Instance variables
        self.data = pd.DataFrame()
        self.metadata = dict()
        self.meaning = dict()

        if metadata is not None:
            self.metadata = metadata
        if meaning is not None:
            self.meaning = meaning
        if df is not None:
            self.from_dataframe(df)
        if path:
            done = self.from_json(path)
            if not done:
                parts = path.split(".")
                if parts[-1] == "nc":
                    # It is a NetCDF file
                    self.from_netcdf(path)
                if parts[-1] == "csv":
                    # It is a CSV file
                    self.from_csv(path)

        # Check if data is multiindex and change to single index
        if isinstance(self.data.index, pd.core.index.MultiIndex):
            try:
                self.data.reset_index(inplace=True)
                self.data.set_index('TIME', inplace=True)
                self.data.index = pd.to_datetime(self.data.index)
                self.data.sort_index(inplace=True)
            except e:
                pass

    def __repr__(self):
        """
        Return a string containing a printable representation of an object.
        """
        parameters = "Parameters: {}".format((", ").join(self.parameters()))

        # Memory use message
        memory_usage = self.memory_usage()
        units = "Bytes"
        if memory_usage > 1000000000:
            memory_usage /= 1000000000
            units = "GBytes"
        elif memory_usage > 1000000:
            memory_usage /= 1000000
            units = "MBytes"
        elif memory_usage > 1000:
            memory_usage /= 1000
            units = "KBytes"
        size_message = "Memory usage: {:.2f} {}".format(memory_usage, units)

        qc_values = self.info_qc()

        # Parameters message
        parameters = self.parameters()
        if not parameters:
            parameters_message = "There is no data."
        else:
            parameters_message = "Parameters:"
            for parameter in parameters:
                try:
                    parameters_message += "\n  - {}: {} ({})".format(
                        parameter, self.meaning[parameter]["long_name"],
                        self.meaning[parameter]["units"])
                except KeyError:
                    parameters_message += \
                        "\n  - {}: Parameter without meaning".format(parameter)

                # Min, max and mean info
                date_min, value_min = self.min(parameter)
                date_max, value_max = self.max(parameter)
                value_mean = self.mean(parameter)
                parameters_message += "\n    - Min value: {:.3f}".format(
                    value_min)
                parameters_message += "\n    - Date min value: {}".format(
                    date_min)
                parameters_message += "\n    - Max value: {:.3f}".format(
                    value_max)
                parameters_message += "\n    - Date max value: {}".format(
                    date_max)
                parameters_message += "\n    - Mean value: {:.3f}".format(
                    value_mean)

                # Percentage of data with QC Flag = 1
                # Calculation of the total of the vales
                total_values = 0
                for _qc_flags, counts in qc_values[parameter].items():
                    total_values += counts
                # Calculation of the percentage
                percentage = 0
                if 1.0 in qc_values[parameter]:
                    percentage = qc_values[parameter][1.0]/total_values*100
                parameters_message += \
                    "\n    - Values with QC = 1: {:.3f} %".format(percentage)

        message = size_message + "\n\n" + parameters_message

        return message

    def from_netcdf(self, path):
        """Load and decode a dataset from a NetCDF file. The compatible
        netCDF files are from the mooring-buoys of
        [EMODNET](http://www.emodnet-physics.eu/Map/),
        [JERICO](http://www.jerico-ri.eu/data-access/),
        [EMSO](http://emso.eu)
        and time series with [NetCDF](http://www.oceansites.org/data/)
        format.

        Parameters
        ----------
            path: str, obj
                Path to a netCDF file or an OpenDAP URL.
                File-like objects are opened with scipy.io.netcdf
                 (onlynetCDF3 supported).
        Returns
        -------
            True/False: Bool
                It indicates if the procedure was successful."""
        def drop(ds_in):
            """Drop some parameters of the dataset.

            Parameters
            ----------
                ds_in: xarray.Dataset
                    Input dataset.
            Returns
            -------
                ds_out: xarray.Dataset
                    Output dataset"""
            # Creation of a list with the variables to drop. We are going to
            # drop all variables with a "_DM" in the key name.
            vars2drop = []
            for k in ds_in.variables.keys():
                if '_DM' in k:
                    vars2drop.append(k)

            # Dropping the previous list and some other variables
            ds_out = ds_in.drop(vars2drop)
            if 'POSITIONING_SYSTEM' in ds_in.variables.keys():
                ds_out = ds_out.drop('POSITIONING_SYSTEM')
            if 'DC_REFERENCE' in ds_in.variables.keys():
                ds_out = ds_out.drop('DC_REFERENCE')
            if 'LATITUDE' in ds_in.variables.keys():
                ds_out = ds_out.drop('LATITUDE')
            if 'LONGITUDE' in ds_in.variables.keys():
                ds_out = ds_out.drop('LONGITUDE')
            if 'POSITION_QC' in ds_in.variables.keys():
                ds_out = ds_out.drop('POSITION_QC')
            if 'DEPH' in ds_in.variables.keys():
                ds_out = ds_out.drop('DEPH')
            if 'DEPH_QC' in ds_in.variables.keys():
                ds_out = ds_out.drop('DEPH_QC')

            return ds_out

        # Check if path contanins a ".nc" file and then, complete the process
        if (isinstance(path, str) and '.nc' in path) or isinstance(path, io.BytesIO):
            # Convert the nc to a xarray dataset
            ds = xr.open_dataset(path)
            # Save metadata
            self.metadata = dict(ds.attrs)
            # Delete not used vars
            ds = drop(ds)
            # Save the meanings
            for variable in ds.variables:
                if '_QC' in variable:
                    continue
                else:
                    self.meaning[variable] = dict(ds[variable].attrs)
            # We add the meaning of DEPTH
            self.meaning['DEPTH'] = {
                "long_name": "depth_of_measure",
                "units": "meters"}

            # Conversion to the dataframe
            self.data = ds.to_dataframe()

            # Set index
            # self.data = self.data.reset_index().set_index('TIME')
            # self.data.sort_index()

            return True
        else:
            return False

    def to_netcdf(self, path):
        """
        It creates a netCDF3 file.

        Parameters
        ----------
            path: str,
                Path to a netCDF file to save.

        Returns
        -------
            True: bool
                It indicates that the process was successfully completed.
        """
        # Sometimes the metadata contains a list of str
        # It can crash the creation of the NetCDF3_64Bits.
        # We are going to change the list of str to str
        metadata_netcdf = self.metadata.copy()
        for key, value in metadata_netcdf.items():
            if isinstance(value, list):
                metadata_netcdf[key] = ', '.join(value)

        # Creation of an xarray dataset
        ds = xr.Dataset(data_vars=self.data, attrs=metadata_netcdf)
        for key in self.parameters():
            ds[key].attrs = self.meaning[key]

        # Creation of the nc file
        ds.to_netcdf(path, format="NETCDF3_64BIT")

        return True

    def from_csv(self, path, metadata=None, meaning=None, **kwds):
        """
        It reads data from a CSV file.

        It uses the pandas.read_csv(). All parameters of read_csv() can be input here.

        Parameters
        ----------
            path: str
                Path of the CSV file.
            metadata: dict
                Metadata dictionary.
            meaning: dict
                Meaning dictionary.
            **kwds: arguments
                All arguments of pandas.read_csv()

        Return
        ------
            True: bool
                Operation successful.
        """
        if metadata is not None:
            self.metadata = metadata
        if meaning is not None:
            self.meaning = meaning

        df = pd.read_csv(path, **kwds)

        self.from_dataframe(df)

        return True

    def from_pickle(self, path):
        """
        Load and decode a WaterFrame object from a pickle file.

        Parameters
        ----------
            path: str
                Path to a pickle file.
        Returns
        -------
            True/False; Bool
                It indicates if the procedure was successful."""
        # Check if path contanins a ".pkl" file and then, complete the process
        if '.pkl' in path:
            temp_dict = pickle.load(open(path, "rb"))
            self.__dict__.clear()
            self.__dict__.update(temp_dict)
            return True
        else:
            return False

    def to_pickle(self, path):
        """It creates a pickle (serialize) file of the WaterFrame.

        Parameters
        ----------
            path: str
                Path to save the pickle file.

        Returns
            True: bool
                If the file is created, it returns True.
        """
        pickle.dump(self.__dict__, open(path, "wb"))

        return True

    def from_dataframe(self, df, metadata=None, meaning=None):
        """
        It creates save the input pandas DataFrame into the WaterFrame.

        Parameters
        ----------
            df: pandas DataFrame
                pandas DataFrame
            metadata : dict
                Metadata information.
            meaning: dict
                Parameter meanings.

        Return
        ------
            True: bool
                Operation successful.
        """
        # Save metadata
        if metadata is not None:
            self.metadata = metadata

        # Save meaning
        if meaning is not None:
            self.meaning = meaning

        # Save df
        self.data = df.copy()
        # Adding QC keys
        df_keys = list(df.keys())
        wf_keys = list(self.data.keys())
        for key in df_keys:
            if "{}_QC".format(key) not in wf_keys:
                self.data["{}_QC".format(key)] = 0

        return True

    def to_csv(self, path):
        """
        It saves the Waterframe data and metadata into a CSV file.

        Parameters
        ----------
            path: str
                Path to save the csv file.

        Returns
        -------
            True/False: bool
                It indicates if the process was successful
        """
        # Create the file
        csv_file = open(path, 'w')

        # Save the metadata
        csv_file.write("#Global attributes;Value\n")
        for key, val in self.metadata.items():
            if isinstance(val, list):
                try:
                    value = ', '.join(val)
                except TypeError:
                    # It happens with data from Pangea
                    # with the "event" key, that it is a list of dict
                    value = ""
            else:
                value = val
            csv_file.write("# ")
            csv_file.write(key)
            csv_file.write(';="')
            csv_file.write(value)
            csv_file.write('"\n')

        # Add three empty lines
        csv_file.write("\n\n\n")

        # Save the data
        self.data.to_csv(csv_file, sep=";", header=True)

        # Close the file
        csv_file.close()

        return True

    def tsplot(self, keys=None, rolling=None, ax=None, average_time=None,
               secondary_y=False, color=None):
        """Plot timeseries.

        Parameters
        ----------
            keys: list of str, str, optional (keys = None)
                keys of self.data to plot. If keys is None, all parameters
                will be ploted.
            rolling: int, optional (rolling = None)
                Size of the moving window. It is the number of observations
                used for calculating the statistic.
            ax: matplotlib.axes object, optional (ax = None)
                It is used to add the plot to an input axes object.
            average_time: str, optional (average_time = None)
                It calculates an average value of a time interval. You can find
                all of the resample options here:
                http://pandas.pydata.org/pandas-docs/stable/timeseries.html#offset-aliases
            secondary_y: bool, optional (secondary_y = False)
                Plot on the secondary y-axis.
            color: str or list of str, optional (color = None)
                Any matplotlib color. It will be applied to the traces.
        Returns
        -------
            ax: matplotlib.AxesSubplot
                New axes of the plot.
        """
        def make_plot(df_in, ax_in, key_in, color_in):
            # Calculation of the std and the mean
            roll = df_in[key_in].rolling(rolling, center=True)
            m = roll.agg(['mean', 'std'])
            # rename 'mean' column
            m.rename(columns={'mean': key_in}, inplace=True)
            ax_out = m[key_in].plot(ax=ax_in, secondary_y=secondary_y,
                                    legend=True, color=color_in)
            ax_out.fill_between(m.index,
                                m[key_in] - m['std'], m[key_in] + m['std'],
                                alpha=.25, color=color_in)
            # Write axes
            try:
                ax_out.set_ylabel(self.meaning[key_in]['units'])
            except KeyError:
                warnings.warn("No units on y_label")

            return ax_out

        # Check keys
        if keys is None:
            keys = self.parameters()

        # Extract data
        df = self.data[keys].dropna().reset_index().set_index('TIME')
        df.index.rename("Date", inplace=True)

        # Resample data
        if average_time is None:
            pass
        else:
            df = df.resample(average_time).mean()

        # Calculation of the rolling value
        if rolling is None:
            if df.size <= 100:
                rolling = 1
            elif df.size <= 1000:
                rolling = df.size//10
            elif df.size <= 10000:
                rolling = df.size // 100
            else:
                rolling = df.size // 1000

        if isinstance(keys, str):
            key = keys
            ax = make_plot(df, ax, key, color)
        elif isinstance(keys, list):
            if color is None:
                for key in keys:
                    ax = make_plot(df, ax, key, color)
            else:
                for key, color_line in zip(keys, color):
                    ax = make_plot(df, ax, key, color_line)
        return ax

    def barplot(self, key, ax=None, average_time=None):
        """Bar Plot.

        Parameters
        ----------
            key: list of str
                keys of self.data to plot.
            ax: matplotlib.axes object, optional (ax = None)
                It is used to add the plot to an input axes object.
            average_time: str, optional (average_time = None)
                It calculates an average value of a time interval. You can find
                all of the resample options here:
                http://pandas.pydata.org/pandas-docs/stable/timeseries.html#offset-aliases
        Returns
        -------
            ax: matplotlib.AxesSubplot
                New axes of the plot.
        """
        def format_year(x):
            return datetime.datetime.\
                strptime(x, '%Y-%m-%d %H:%M:%S').strftime('%Y')

        # Extract data
        df = self.data[key].dropna().reset_index().set_index('TIME')
        df.index.rename("Date", inplace=True)

        # Resample data
        if average_time is None:
            pass
        else:
            df = df.resample(average_time).mean()

        if isinstance(key, list):
            ax = df[key].plot.bar(ax=ax, legend=True)
        else:
            ax = df[key].plot.bar(ax=ax)
            # Write axes
            try:
                ax.set_ylabel(self.meaning[key]['units'])
            except KeyError:
                print("Warning: We don't know the units of", key,
                      "Please, add info into self.meaning[", key, "['units']")

            if average_time == 'A':
                ax.set_xticklabels([format_year(x.get_text())
                                    for x in ax.get_xticklabels()])

        return ax

    def hist(self, parameter=None, mean_line=False, **kwds):
        """Make a histogram of the WaterFrame's.

        A histogram is a representation of the distribution of data.
        This function calls DataFrame.hist(), on each parameter of the
        WaterFrame, resulting in one histogram per parameter.

        Parameters
        ----------
            parameter: str, list of str, optional (parameter=None)
                keys of self.data to plot. If parameter=None, it will plot all
                parameters.
            mean_line: bool, optional (mean_line=False)
                 It draws a line representing the average of the dataset.
            **kwds:
                All other plotting keyword arguments to be passed to
                DataFrame.hist().
                https://pandas.pydata.org/pandas-docs/version/0.23/generated/pandas.DataFrame.hist.html
        Returns
        -------
            axes: matplotlib.AxesSubplot or numpy.ndarray of them
                New axes of the plot.
        """
        if parameter is None:
            parameter = self.parameters()

        if isinstance(parameter, str):
            parameter = [parameter]

        axes = self.data.hist(column=parameter, **kwds)
        parameter_counter = 0
        try:
            for ax in axes:
                ax.set_xlabel("Values")
                ax.set_ylabel("Frequency")
                if mean_line is True:
                    if parameter_counter < len(parameter):
                        x_mean = self.mean(parameter[parameter_counter])
                        ax.axvline(x_mean, color='k',
                                   linestyle='dashed',
                                   linewidth=1)

                        parameter_counter += 1
        except AttributeError:

            # Creation of the mean line
            parameter_counter = 0
            for irow in range(len(axes)):
                for icol in range(len(axes[irow])):
                    if parameter_counter < len(parameter):
                        axes[irow, icol].set_xlabel("Values")
                        axes[irow, icol].set_ylabel("Frequency")
                        if mean_line is True:
                            x_mean = self.mean(axes[irow, icol].get_title())
                            axes[irow, icol].axvline(x_mean, color='k',
                                                     linestyle='dashed',
                                                     linewidth=1)
                        parameter_counter += 1

        return axes

    def scatter_matrix(self, keys, ax=None):
        """
        Draw a matrix of scatter plots.

        Parameters
        ----------
            key: list of str
                keys of self.data to plot.
                Keys must contain different words.
                ex:
                    keys = ['VAVH', 'VCMX'] is ok.
                    keys = ['VAVH', 'VAVH'] is not ok.
            ax: matplotlib.axes object, optional (ax = None)
                It is used to add the plot to an input axes object.

        Returns
        -------
            ax: matplotlib.AxesSubplot or False
                New axes of the plot.
                It returns False if operation is not successful.
        """
        if isinstance(keys, list):
            # Remove duplicate keys
            keys = list(set(keys))
            if len(keys) > 1:
                # Extract data
                df = self.data[keys]
                ax = pd.plotting.scatter_matrix(df, diagonal='kde', ax=ax)
                return ax
            else:
                return False
        else:
            return False

    def qcplot(self, parameters=None, ax=None):
        """
        Plot the timeserie with dots of different colours according to the QC
        Flag.

        Parameters
        ----------
            parameters: string or list of strings, optional
            (parameters = None)
                key of self.data to apply the test.
            ax: matplotlib.axes object, optional (ax = None)
                It is used to add the plot to an input axes object.
        Returns
        -------
            ax: matplotlib.AxesSubplot or False
                New axes of the plot. It returns False if the operation was
                not successful.
        """
        if parameters is None:
            parameters = self.parameters()
        elif isinstance(parameters, str):
            parameters = [parameters]
        elif isinstance(parameters, list):
            pass
        else:
            return False

        for parameter in parameters:
            if "_QC" in parameter:
                return False
            else:
                # Extract data
                df = self.data[
                    [parameter, parameter+'_QC']].dropna().reset_index().set_index('TIME')
                df.index.rename("Date", inplace=True)
                # Sometimes Depth is on the df
                if "DEPTH" in df.keys():
                    del df['DEPTH']
                # Create a dataframe divided with flags
                for flag in range(0, 10):
                    df_ = pd.DataFrame()
                    df_['{}'.format(flag)] = df.ix[df[parameter+'_QC'] == flag, parameter]
                    if df_.index.size > 0:
                        # Change line style -> Maybe not necessary
                        line = ""
                        ax = df_.plot(ax=ax, marker='.', linestyle=line)
                ax.set_ylabel(self.meaning[parameter]['units'])
                ax.legend(title="QC Flags")

        return ax

    def qcbarplot(self, parameters=None, ax=None):
        """
        Bar plot with the count of values with a  '_QC' in the keys of
        self.data.

        Parameters
        ----------
            parameters: string or list of strings, optional
            (parameters = None)
                key of self.data to apply the test.
            ax: matplotlib.axes object, optional (ax = None)
                It is used to add the plot to an input axes object.
        Returns
        -------
            ax: matplotlib.AxesSubplot or False
                New axes of the plot. It returns False if the operation was
                not successful.
        """
        if parameters is None:
            parameters = self.parameters()
        elif isinstance(parameters, str):
            parameters = [parameters]
        elif isinstance(parameters, list):
            pass
        else:
            return False

        key_list = []  # List of parameter names
        values_list = []  # List of counters of each flag
        for parameter in parameters:
            if "_QC" in parameter:
                return False
            else:
                key_list.append(parameter)
                values_list.append(pd.value_counts(self.data[parameter+"_QC"]))

            # Obtain the different values of QC
            qc_values = []
            for value in values_list:
                for key in value.keys():
                    if key not in qc_values:
                        qc_values.append(key)

            # Creation of a dataframe with the count info
            df = pd.DataFrame(values_list, key_list)
            # Creation of the plot
            ax = df.sort_index().plot.bar(ax=ax, legend=False)
            # Set labels
        ax.set_ylabel("Number of measurements")
        #  TODO: Bug on labels
        ax.legend(title="QC Flags", labels=qc_values)

        return ax

    def spectroplot(self):
        """
        It plots the spectrometer of the acoustic data.

        Returns
        -------
            matplotlib.Figure
                New axes of the plot.
        """

        # Prepare data
        time = self.data.index
        wavelength = []
        values = []
        for key in self.data.keys():
            try:
                wavelength.append(float(key))
                values.append(self.data[key].values)
            except ValueError:
                pass

        # Creation of figure
        return plt.pcolormesh(time, wavelength, values)

    def profileplot(self, parameter_y, parameter_x=None, ax=None,):
        """
        It creates a graph a profile plot.
        Y-axes suppose to be a depth related parameter.

        Parameters
        ----------
            parameter_y: str
                y-axes parameter.
            parameters_x: list of str, str (optional,
            parameters_x = None)
                x-axes parameter.
            ax: matplotlib.axes object, optional (ax = None)
                It is used to add the plot to an input axes object.
        Returns
        -------
            ax_out: matplotlib.AxesSubplot
                New axes of the plot.
        """

        # Check the type of parameter_x
        if isinstance(parameter_x, str):
            parameter_x = [parameter_x]

        # Extract data
        parameter_y = [parameter_y]
        keys = parameter_y + parameter_x
        df = self.data[keys].dropna().reset_index().set_index('TIME')
        df.index.rename("Date", inplace=True)
        
        ax_out = df.plot.scatter(x=parameter_x[0], y=parameter_y[0], ax=ax)

        if len(parameter_x) == 1:
            ax_out.set_xlabel("{} ({})".format(self.meaning[parameter_x[0]]['long_name'],
                                               self.meaning[parameter_x[0]]['units']))
        ax_out.set_ylabel("{} ({})".format(self.meaning[parameter_y[0]]['long_name'],
                                           self.meaning[parameter_y[0]]['units']))

        ax_out.invert_yaxis()
        ax_out.ticklabel_format(useOffset=False)
        # ax_out.legend().set_visible(False)
        ax_out.set_title('Profile Plot')

        return ax_out

    def spike_test(self, parameters=None, window=0, threshold=3, flag=4):
        """
        It checks if there is any spike in the time series.

        Parameters
        ----------
            parameters: string or list of strings, optional
            (parameters = None)
                key of self.data to apply the test.
            window: int, optional (window = 0)
                Size of the moving window of values to calculate the mean.
                If it is 0, the function calculates the optimal window.
            threshold: int, optional (threshold = 3)
                The z-score at which the algorithm signals.
            flag: int, optional (flag = 4)
                Flag value to write in on the fail values.

        Returns
        -------
            True/False: bool
                It indicates if the process is (not) successful.
        """

        if parameters is None:
            parameters = self.parameters()
        elif isinstance(parameters, str):
            parameters = [parameters]
        elif isinstance(parameters, list):
            pass
        else:
            return False

        for parameter in parameters:
            # Auto calculation of window
            if window == 0:
                window = int(len(self.data[parameter])/100)
                if window < 3:
                    window = 3
                elif window > 100:
                    window = 100

            signals = self.data[
                parameter].rolling(window=window, center=True).mean().fillna(method='bfill')
            difference = np.abs(self.data[parameter] - signals)
            outlier_idx = difference > threshold
            self.data.ix[outlier_idx, parameter + '_QC'] = flag

        return True

    def range_test(self, parameters=None, flag=4, limits=None):
        """
        Check if the values of a parameter are out of range.

        Parameters
        ----------
            parameters: string or list of strings, optional
            (parameters = None)
                key of self.data to apply the test.
            flag: int, optional (flag = 4)
                Flag value to write in on the fail values.
            limits: tuple, optional (range = None)
                (Min value, max value) of the range of correct values.

        Returns
        -------
            True/False: bool
                It indicates if the process is (not) successful.
        """
        ranges = {
            'ATMP': [600, 1500],  # atmospheric pressure at altitude
            'ATMS': [0, 2000],  # Atmospheric pressure at sea level
            'CHLT': [0, 30],  # total chlorophyll
            'CNDC': [0, 30],  # electrical conductivity
            'GSPD': [0, 40],  # gust wind speed
            'HCDT': [0, 360],  # current to direction relative true north
            'HEAD': [0, 360],  # PLAT. HEADING REL. TRUE NORTH
            'LINC': [0, 3000],  # long-wave incoming radiation
            'LW': [0, 50],  # Downwelling vector radiance as energy
            'OSAT': [0, 200],  # oxygen saturation
            'PHPH': [0, 14],  # ph
            'PRES': [0, 500],  # sea pressure
            'PRRT': [0, 200],  # hourly precipitation rate
            'PSAL': [0, 40],  # practical salinity
            'RDIN': [0, 1100],  # incident radiation
            'SVEL': [130, 180000],  # sound velocity
            'SWDR': [0, 360],  # SWELL DIRECTION REL TRUE N.
            'SWHT': [0, 50],  # SWELL HEIGHT
            'SWPR': [0, 20],  # SWELL PERIOD
            'TEMP': [0, 50],  # Sea temperature
            'VAVH': [0, 20],  # AVER. HEIGHT HIGHEST 1/3 WAVE
            'VAVT': [0, 20],  # AVER. PERIOD HIGHEST 1/3 WAVE
            'VCMX': [0, 20],  # MAX CREST TROUGH WAVE HEIGHT
            'VDIR': [0, 360],  # wave direction rel. true north
            'VEPK': [0, 100],  # WAVE SPECTRUM PEAK ENERGY
            'VHM0': [0, 20],  # SPECTRAL SIGNIFICANT WAVE HEIGHT
            'VMDR': [0, 360],  # Mean wave direction
            'VPED': [0, 360],  # dir. spreading at wave peak
            'VPSP': [0, 360],  # dir. spreading at wave peak
            'VSMC': [0, 20],  # SPECTUM MOMENT(0, 2) WAVE PERIOD
            'VTDH': [0, 40],  # significant wave height
            'VTM02': [0, 13],  # Spectral moments (0, 2) wave period (Tm02)
            'VTPK': [0, 40],  # WAVE SPECTRUM PEAK PERIOD
            'VTZA': [0, 40],  # AVER ZERO CROSSING WAVE PERIOD
            'VTZM': [0, 40],  # period of the highest wave
            'VZMX': [0, 20],  # Maximum zero crossing wave height
            'WDIR': [0, 360],  # Wind from direction relative true north
            'WSPD': [0, 50],  # Horizontal wind speed
        }

        if parameters is None:
            parameters = self.parameters()
        elif isinstance(parameters, str):
            parameters = [parameters]
        elif isinstance(parameters, list):
            pass
        else:
            return False

        for parameter in parameters:
            if limits:
                self.data.ix[self.data[parameter] < limits[0], parameter + '_QC'] = flag
                self.data.ix[self.data[parameter] > limits[1], parameter + '_QC'] = flag
            elif parameter in ranges.keys():
                self.data.ix[self.data[parameter] < ranges[parameter][0], parameter + '_QC'] = flag
                self.data.ix[self.data[parameter] > ranges[parameter][1], parameter + '_QC'] = flag
            else:
                return False
        return True

    def flat_test(self, parameters=None, window=2, flag=4):
        """
        It detects if there are equal consecutive values in the time series.

        Parameters
        ----------
            parameters: string or list of strings, optional
            (parameters = None)
                key of self.data to apply the test.
            window: int, optional (window = 1)
                Size of the moving window of values to calculate the mean.
                If it is 0, the function calculates the optimal window.
            flag: int, optional (flag = 4)
                Flag value to write in on the fail values.

        Returns
        -------
            True/False: bool
                It indicates if the process is (not) successful.
        """
        if parameters is None:
            parameters = self.parameters()
        elif isinstance(parameters, str):
            parameters = [parameters]
        elif isinstance(parameters, list):
            pass
        else:
            return False

        if window == 0:
            window = 2
        if window > 0:
            df = self.data.rolling(window).std()

        for parameter in parameters:
            if '_QC' in parameter:
                return False
            else:
                self.data.ix[df[parameter] == 0, parameter + '_QC'] = flag
                return True

    def flag2flag(self, parameters=None, original_flag=0, translated_flag=1):
        """
        It changes the flags of the key, from original_flag to translated_flag.

        Parameters
        ----------
            parameters: string, list of strings optional
            (parameters = None)
                Key of self.data to apply the test.
            original_flag: int, optional (original_flag = 0)
                Flag number to translate.
            translated_flag: int, optional (translated_flag = 1)
                Translation of the original flag number.

        Returns
        -------
            True/False: bool
                The operation is (not) successful.
        """
        if parameters is None:
            parameters = self.parameters()
        elif isinstance(parameters, str):
            parameters = [parameters]
        elif isinstance(parameters, list):
            pass
        else:
            return False

        for parameter in parameters:
            if '_QC' in parameter:
                return False
            self.data[parameter+'_QC'].replace(original_flag, translated_flag,
                                               inplace=True)

    def reset_flag(self, parameters=None, flag=0):
        """
        It sets the flag values of the parameter to "flag".

        Parameters
        ----------
            parameters: string, list of strings optional
            (parameters = None)
                Key of self.data to apply the test.
            flag: int, optional (flag = 0)
                Flag value to write.

        Returns
        -------
            True/False: bool
                The operation is (not) successful.
        """
        if parameters is None:
            parameters = self.parameters()
        elif isinstance(parameters, str):
            parameters = [parameters]
        elif isinstance(parameters, list):
            pass
        else:
            return False

        for parameter in parameters:
            if parameter in self.parameters():
                self.data[parameter + '_QC'] = flag
            else:
                return False
        return True

    def qc(self, parameters=None, window=0, threshold=3, bad_flag=4, good_flag=1):
        """
        Auto QC process.

        Parameters
        ----------
            parameters: string, list of strings optional
            (parameters = None)
                Key of self.data to apply the test.
            window: int, optional (window = 0)
                Size of the moving window of values to calculate the mean.
                If it is 0, the function calculates the optimal window.
            threshold: int, optional (threshold = 3)
                The z-score at which the algorithm signals.
            bad_flag: int, optional (flag = 4)
                Flag value to write in on the fail values.
            good_flag: int, optional (flag = 1)
                Flag value to write in on the good values.

        Returns
        -------
            True/False: bool
                The operation is (not) successful.
        """

        if parameters is None:
            parameters = self.parameters()
        elif isinstance(parameters, str):
            parameters = [parameters]
        elif isinstance(parameters, list):
            pass
        else:
            return False

        self.reset_flag(parameters)
        self.range_test(parameters, flag=bad_flag)
        self.spike_test(parameters, window=window, threshold=threshold, flag=bad_flag)
        self.flat_test(parameters, window=window, flag=bad_flag)
        answer = self.flag2flag(parameters, translated_flag=good_flag)

        return answer

    def info_qc(self):
        """
        It returns a dictionary with the count and percentage of the QC flag.

        Return
        ------
            qc_dict: dict
                {NAME OF THE PARAMETER: {QC FLAG: int}}
        """
        qc_dict = {}
        for parameter in self.parameters():
            info = self.data[parameter+"_QC"].value_counts().to_dict()
            qc_dict[parameter] = info

        return qc_dict

    def drop(self, keys, flags=None):
        """
        Remove required keys (and associated QC keys) from self.data requested
        axis removed.

        Parameters
        ----------
            keys: str, list of str
                keys of self.data to drop.
            flags: list of int, int, None, optional (flags = None)
                Number of flag to drop. It can be None, int or a list of int.
                If it is None, column will be deleted.
        Returns
        -------
            True/False: bool
                It indicates that the process was successful."""
        # Look for QC keys and append to "keys"
        keys_qc = []
        if isinstance(keys, str):
            key_qc = keys + "_QC"
            keys_qc.append(key_qc)
            # Cast keys (str) to list
            keys = [keys]
            keys += keys_qc
        elif isinstance(keys, list):
            for key in keys:
                # Check if key+"_QC" exists in self.data
                if key + "_QC" in self.data.keys():
                    keys_qc.append(key+"_QC")
            keys += keys_qc
        # Drop
        if flags is None:
            self.data.drop(keys, axis=1, inplace=True)
        else:
            # If it is an Int, cast to list
            if isinstance(flags, int):
                flags = [flags]
            for key_qc in keys_qc:
                for flag in flags:
                    try:
                        self.data.drop(
                            self.data[self.data[key_qc] == flag].index,
                            inplace=True)
                    except KeyError:
                        # No data with this QC
                        pass
        return True

    def rename(self, old_name, new_name):
        """
        It renames keys of self.data.

        Parameters
        ----------
            old_name: str
                key name to change.
            new_name: str
                New name of the key.

        Returns
        -------
            True: bool
                The operation is successful.

        """
        self.data.rename(columns={old_name: new_name}, inplace=True)
        self.data.rename(columns={old_name+'_QC': new_name+'_QC'},
                         inplace=True)
        try:
            self.meaning[new_name] = self.meaning.pop(old_name)
            return True
        except KeyError:
            warnings.warn("Parameter without meaning")

    def concat(self, waterframe):
        """
        The concat function does all of the heavy lifting of performing
        concatenation operations along an axis while performing optional set
        logic (union or intersection) of the indexes on the other axes.

        Parameters
        ----------
            waterframe: WaterFrame
                WaterFrame object to concat to self

        Returns
        -------
            True: bool
                The operation is successful.
        """
        if self.data.index.size > 0:
            # Check if keys of waterframes contain the same name
            keys_self = self.data.keys()
            keys_new = waterframe.data.keys()
            for key in keys_new:
                if '_QC' in key:
                    continue
                if key in keys_self:
                    # It is a repeated key

                    # Look for number of copy
                    copy_number = 1
                    saved_number = 0
                    for key_self in keys_self:
                        if '_QC' in key_self:
                            continue
                        if '{}(NEW'.format(key) in key_self:
                            saved_number = int(
                                key_self.split("(NEW")[1].split(")")[0])
                            saved_number += 1
                    if saved_number > copy_number:
                        copy_number = saved_number

                    # Rename the new waterframe
                    waterframe.rename(key,
                                      "{}(NEW{})".format(key, copy_number))
                    keys_new += ['{}(SN{})'.format(key, copy_number)]

            # Delete duplicated indexes
            self.data = self.data.loc[
                ~self.data.index.duplicated(keep='first')]
            waterframe.data = waterframe.data.loc[
                ~waterframe.data.index.duplicated(keep='first')]

            # Concat dataframes
            frames = [self.data, waterframe.data]
            self.data = pd.concat(frames, axis=1)

            # Merge dictionaries
            self.meaning = {**self.meaning, **waterframe.meaning}
            # Merge of metadata
            for key in waterframe.metadata.keys():
                if key in self.metadata.keys():
                    if waterframe.metadata[key] == self.metadata[key]:
                        # Skip if we already have the metadata.
                        continue
                    else:
                        # Merge the two values of metadata.
                        self.metadata[key] += ", {}".format(
                            waterframe.metadata[key])

        else:
            self.data = waterframe.data.copy()
            self.meaning = waterframe.meaning.copy()
            self.metadata = waterframe.metadata.copy()

        return True

    def resample(self, rule, method='mean'):
        """
        Convenience method for frequency conversion and sampling of
        time series of the WaterFrame object.

        Parameters
        ----------
            rule: str
                The offset string or object representing target conversion.
                You can find all of the resample options here:
                http://pandas.pydata.org/pandas-docs/stable/timeseries.html#offset-aliases
            method: "mean", "max", "min", optional, (method = "mean")
                Save the new value with the mean(), max() or min() function.

        Returns
        -------
            True/False: bool
                It indicates that the process was successful.
        """
        # It only can be possible if the dataframe only contain 1 index.
        if len(self.data.index.names) == 1:
            if method == "mean":
                self.data = self.data.resample(rule).mean()
            elif method == "max":
                self.data = self.data.resample(rule).max()
            elif method == "min":
                self.data = self.data.resample(rule).min()
            # Change "_QC" values to 0
            for key in self.data.keys():
                if "_QC" in key:
                    self.data[key] = 0
            return True
        else:
            return False

    def slice_time(self, start=None, end=None):
        """
        Delete data outside the time interval.

        Parameters
        ----------
            start: str, timestamp, optional (start=None)
                Start time interval with format 'YYYYMMDDhhmmss' or timestamp.
            end: str, timestamp, optional (end=None)
                End time interval with format 'YYYYMMDDhhmmss' or timestamp.

        Returns
        -------
            True: bool
                The operation is successful.

        """
        if isinstance(self.data.index, pd.core.index.MultiIndex):
            # Sometimes, you have multiindex TIME and DEPTH
            self.data.reset_index(inplace=True)
            self.data.set_index('TIME', inplace=True)

        self.data.sort_index(inplace=True)

        if start is None:
            datetime_end = datetime.datetime.strptime(end, '%Y%m%d%H%M%S')
            end_slice = self.data.index.searchsorted(datetime_end)
            self.data = self.data.ix[:end_slice]

        if end is None:
            datetime_start = datetime.datetime.strptime(start, '%Y%m%d%H%M%S')
            start_slice = self.data.index.searchsorted(datetime_start)
            self.data = self.data.ix[start_slice:]

        if start is not None and end is not None:
            datetime_start = datetime.datetime.strptime(start, '%Y%m%d%H%M%S')
            start_slice = self.data.index.searchsorted(datetime_start)
            datetime_end = datetime.datetime.strptime(end, '%Y%m%d%H%M%S')
            end_slice = self.data.index.searchsorted(datetime_end)
            self.data = self.data.ix[start_slice:end_slice]

        return True

    def clear(self):
        """
        It delete all data and metadata of the waterframe.
        """
        self.data = pd.DataFrame()
        self.metadata = dict()
        self.meaning = dict()

    def memory_usage(self):
        """
        Memory usage of the WaterFrame.

        Returns
        -------
            size: int
                Number of bytes in use.
        """
        size = sys.getsizeof(self.data) + sys.getsizeof(self.metadata)
        return size

    def parameters(self):
        """
        It return the parameter list used in this WaterFrame.
        The parameters are the keys of self.data without "_QC".

        Returns
        -------
            parameter_list: list of str
                Parameters used in the WaterFrame.
        """

        parameter_list = [key for key in self.data.keys()
                          if "_QC" not in key
                          if key+"_QC" in self.data.keys()]

        return parameter_list

    def use_only(self, parameters=None, flags=None, dropnan=False):
        """
        Drop all parameters not presented in the input list with QC flags
        different than given in the input flags.

        Parameters
        ----------
            parameters: list of str, str, optional (parameters = None)
                Parameter to save in the WaterFrame.
            flags: list of int, int, None, optional (flags = None)
                QC Flag of the parameter to save.
            dropnan: Bool, optional (dropnan = False)
                Drop all lines of self.data that contain a nan in any of their
                columns.

        Returns
        -------
            True/False: bool
                The operations is successful.

        """

        if parameters is None:
            parameters = self.parameters()

        if isinstance(parameters, str):
            parameters = [parameters]

        # Check if parameters exist
        for parameter in parameters:
            if parameter not in self.parameters():
                warnings.warn("Parameter not found.")
                return False

        # Calculation of parameters to drop
        drop_parameters = [drop_param for drop_param in self.parameters()
                           if drop_param not in parameters]

        if drop_parameters:
            self.drop(keys=drop_parameters)
        if flags:
            if isinstance(flags, int):
                flags = [flags]
            # Calculation of flags to drop
            drop_flags = [i for i in range(10) if i not in flags]
            self.drop(keys=parameters, flags=drop_flags)
        if dropnan:
            self.data.dropna(inplace=True)

        return True

    def corr(self, parameter1, parameter2, method='pearson', min_periods=1):
        """
        Compute pairwise correlation of data columns of parameter1 and
        parameter2, excluding NA/null values.

        Parameters
        ----------
            parameter1: str
                Key name of the column 1 to correlate.
            parameter2: str
                Key name of the column 2 to correlate.
            method: {‘pearson’, ‘kendall’, ‘spearman’}
                pearson : standard correlation coefficient
                kendall : Kendall Tau correlation coefficient
                spearman : Spearman rank correlation
            min_periods : int, optional
                Minimum number of observations required per pair of columns to
                have a valid result. Currently only available for pearson and
                spearman correlation
        Returns
        -------
            correlation_number: float
        """
        correlation_number = self.data[
            parameter1].corr(self.data[parameter2], method=method,
                             min_periods=min_periods)
        return correlation_number

    def max_diff(self, parameter1, parameter2):
        """
        Calculation the maximum difference between values of two parameters.

        Parameters
        ----------
            parameter1: str
                Key name of the column 1 to calculate the difference.
            parameter2: str
                Key name of the column 2 to calculate the difference.
        Returns
        -------
            (where, value): Pandas DataFrame Index, float
                It returns the position (index) and value of the maximum
                difference.
        """

        where = (self.data[parameter1] - self.data[parameter2]).abs().idxmax()
        value = (self.data[parameter1] - self.data[parameter2]).abs().max()

        return (where, value)

    def datetime_intervals(self, parameter):
        """
        It returns the index (TIME) of intervals between NaNs.

        Parameters
        ----------
            parameter: str
                Parameter to check.

        Returns
        -------
            intervals: [(str, str)]
                List of tuples with the start and end time of each interval of
                data.
        """
        # Creation of a timeseries with the positions of null values
        ts = self.data[parameter].isnull()

        # Check where are the intervals
        intervals = []
        in_interval = False
        for index, value in ts.items():
            if in_interval is False and value is False:
                in_interval = True
                start = index.strftime('%Y-%m-%d %H:%M:%S')
            if in_interval is True and value is True:
                in_interval = False
                end = index.strftime('%Y-%m-%d %H:%M:%S')
                intervals.append((start, end))

        return intervals

    def mean(self, parameter):
        """
        It returns the mean value of a parameter.

        Parameters
        ----------
            parameter: str
                Parameter to calculate the mean.

        Returns
        -------
            mean_value:
                Mean of values of the parameter.
        """

        mean_value = self.data[parameter].mean()

        return mean_value

    def max(self, parameter):
        """
        It returns the max value of a parameter.

        Parameters
        ----------
            parameter: str
                Parameter to calculate the mean.

        Returns
        -------
            (where, max_value): (str, float)
                (Index of the max value of the parameter, Max values of the
                parameter)
        """

        where = self.data[parameter].idxmax()
        max_value = self.data[parameter].max()

        return (where, max_value)

    def min(self, parameter):
        """
        It returns the min value of a parameter.

        Parameters
        ----------
            parameter: str
                Parameter to calculate the mean.

        Returns
        -------
            (where, min_value): (str, float)
                (Index of the minimum value, Min values of the parameter).
        """
        where = self.data[parameter].idxmin()
        min_value = self.data[parameter].min()

        return (where, min_value)

    def copy(self):
        """
        It creates a copy of self

        Returns
        -------
            c: WaterFrame
                Copy of self.
        """
        c = copy.deepcopy(self)
        return c

    def info_metadata(self, keys=None):
        """
        It returns a string with the metadata information.

        Parameters
        ----------
            keys: string or list of strings (optional)
                The return message will contain the information of the input keys.
                If keys is None, all keys will be added to the return message.

        Returns
        -------
            message: string
                Message with the metadata information.
        """

        if keys is None:
            keys = self.metadata.keys()

        message = ""
        for key, value in self.metadata.items():
            if key in keys and str(value).strip() != "":
                message += "  - {}: {}\n".format(key, value)

        return message[:-1]

    def info_meaning(self, keys=None):
        """
        It returns a string with the meanings information.

        Parameters
        ----------
            keys: string or list of strings (optional)
                The return message will contain the information of the input keys.
                If keys is None, all keys will be added to the return message.

        Returns
        -------
            message: string
                Message with the meanings information.
        """

        if keys is None:
            keys = self.meaning.keys()

        message = ""
        for key, meaning_dict in self.meaning.items():
            if key in keys:
                message += "  - {}\n".format(key)
                for meaning_key, meaning_value in meaning_dict.items():
                    if meaning_value.strip() != "":
                        message += "    - {}: {}\n".format(meaning_key, meaning_value)

        return message[:-1]

    def __setitem__(self, key, data):
        """
        Set the value of *data* at *key* of *WaterFrame.data*.

        Parameters
        ----------
        key: string or list of string
            Key of DataFrame *WaterFrame.data*.
        data: any
            Values to be placed into WaterFrame.data[key].
        """
        self.data[key] = data

    def __getitem__(self, key):
        """
        Get the values of *WaterFrame.data[key]*.

        Parameters
        ----------
        key: string or list of string
            Key of DataFrame *WaterFrame.data*.
        """
        return self.data[key]

    def plot(self, **kwds):
        """
        It calls the pandas DataFrame.plot() method.

        Parameters
        ----------
            **kwds: arguments
                pandas plot() arguments.

        Returns
        -------
            axes : matplotlib.axes.Axes or numpy.ndarray of them
        """
        axes = self.data.plot(**kwds)

        return axes

    def get_coordinates(self):
        """
        It returns the minimum and maximum coordinates placed in the metadata.

        Returns
        -------
            coordinates: tuple of tuples
                ((minimum latitude, minimum longitude, minimum depth),
                 (maximum latitude, maximum longitude, maximum depth))
        """
        min_lat = None
        min_lon = None
        max_lat = None
        max_lon = None
        min_depth = None
        max_depth = None

        metadata_keys = self.metadata.keys()
        if "geospatial_lat_min" in metadata_keys:
            min_lat = self.metadata["geospatial_lat_min"]
        if "geospatial_lat_max" in metadata_keys:
            max_lat = self.metadata["geospatial_lat_max"]
        if "geospatial_vertical_min" in metadata_keys:
            min_depth = self.metadata["geospatial_vertical_min"]
        if "geospatial_vertical_max" in metadata_keys:
            max_depth = self.metadata["geospatial_vertical_max"]
        if "geospatial_lon_max" in metadata_keys:
            min_lon = self.metadata["geospatial_lon_max"]
        if "geospatial_lon_max" in metadata_keys:
            max_lon = self.metadata["geospatial_lon_max"]

        coordinates = ((min_lat, min_lon, min_depth), (max_lat, max_lon, max_depth))

        return coordinates

    def to_json(self):
        """
        It creates a JSON string with the WaterFrame information.

        Returns
        -------
            json_string: str
                JSON string.
        """

        big_dict = {
            "metadata": self.metadata,
            "meaning": self.meaning,
            "data": self.data.to_json()
        }

        json_string = json.dumps(big_dict)

        return json_string

    def from_json(self, json_string):
        """
        It loads a WaterFrame object from a JSON string.

        Parameters
        ----------
            json_string: str
                String that contains a JSON.

        Returns
        -------
            done: bool
                True if the operation is successful.
        """

        done = False
        try:
            big_dict = json.loads(json_string)

            keys = big_dict.keys()
            if "metadata" in keys:
                self.metadata = big_dict["metadata"].copy()
            if "meaning" in keys:
                self.meaning = big_dict["meaning"].copy()
            if "data" in keys:
                self.data = pd.read_json(big_dict["data"])
            else:
                self.data = pd.read_json(big_dict)

            done = True
        except ValueError:
            pass
        return done

    def __getattr__(self, key):
        """
        Get values from *self.metadata[key]*.

        Parameters
        ----------
            key: str
                key of self.metadata

        Returns
        -------
            value:
                Value of self.metadata
        """

        value = self.metadata[key]

        return value

    def empty(self):
        """
        It return True if the dataframe is empty.

        Returns
        -------
            empty: bool
        """
        empty = self.data.empty
        return empty



"""Module with the class IPlot, designed to create Plotly charts"""
import plotly.graph_objs as go
import pandas as pd


class IFig:
    """ It contains methods to create Plotly charts from WaterFrames"""

    def __init__(self, wf):
        """
        Constructor of the class.

        Parameters
        ----------
            wf: WaterFrame
        """
        self.wf = wf

        # Check if wf is multiindex and change to single index
        if isinstance(self.wf.data.index, pd.core.index.MultiIndex):
            self.wf.data.reset_index(inplace=True)
            self.wf.data.set_index('TIME', inplace=True)

    def site_map(self):
        """
        It creates a chart with a map with the coordinates of the site. It looks for the coordinates
        in the metadata information.

        Returns
        -------
            figure: dict
                Plotly Figure dictionary.
        """
        coordinates = self.wf.get_coordinates()

        # For now we are going to plot the minimum coordinates. It will change in the future.
        lat = coordinates[0][0]
        lon = coordinates[0][1]

        # Data creation
        data = [
            dict(
                type="scattergeo",
                lon=[lon],
                lat=[lat],
                mode='markers',
            )
        ]

        # Configuration of the zoom
        if lat is not None and lon is not None:
            lataxis = dict(range=[float(lat)-9, float(lat)+9])
            lonaxis = dict(range=[float(lon)-16, float(lon)+16])
        else:
            lataxis = None
            lonaxis = None

        layout = dict(
            geo=dict(
                lakecolor="rgb(255, 255, 255)",
                resolution=50,
                showcoastlines=True,
                showland=True,
                landcolor="rgb(229, 229, 229)",
                countrycolor="rgb(255, 255, 255)",
                coastlinecolor="rgb(255, 255, 255)",
                lataxis=lataxis,
                lonaxis=lonaxis
            ),
            margin=dict(l=10, r=10, t=0, b=0),
        )

        figure = dict(data=data, layout=layout)
        return figure

    def time_series(self, parameters=None):
        """
        It creates a Plotly time-series figure.

        Parameters
        ----------
            parameters: str or list (optional, parameters=None)
                Parameters to plot.

        Returns
        -------
            figure: dict
                Plotly figure dictionary
        """
        figure = {'data': None, 'layout': None}

        # Check parameters
        if parameters is None:
            parameters = self.wf.parameters()
        elif isinstance(parameters, str):
            parameters = [parameters]
        elif isinstance(parameters, list):
            pass
        else:
            return figure

        data = [go.Scatter(x=self.wf.data.index, y=self.wf[parameter], fill="tozeroy",
                           name=parameter) for parameter in parameters]

        # Layout
        y_label = None
        if len(parameters) == 1:
            try:
                title = self.wf.meaning[parameters[0]]['long_name']
            except KeyError:
                title = parameters[0]
            try:
                y_label = self.wf.meaning[parameters[0]]['units']
            except KeyError:
                pass
        else:
            title = "_".join(parameters)
        min_value = None
        max_value = None
        for parameter in parameters:

            _min = min(self.wf[parameter])
            _max = max(self.wf[parameter])

            if min_value is None or _min < min_value:
                min_value = _min

            if max_value is None or _max > max_value:
                max_value = _max

        layout = {
            'title': title,
            'yaxis': {
                'range': [min_value, max_value],
                'title': y_label
            },
            'margin': {'l': 50, 'r': 10, 't': 45, 'b': 30}
        }

        figure = {"data": data, "layout": layout}

        return figure

    @staticmethod
    def site_map_from_waterframe(waterframe):
        """
        It creates a chart with a map with the coordinates of the site. It looks for the coordinates
        in the metadata information.

        Parameters
        ----------
            wf: WaterFrame

        Returns
        -------
            figure: dict
                Plotly Figure dictionary.
        """
        coordinates = waterframe.get_coordinates()

        # For now we are going to plot the minimum coordinates. It will change in the future.
        lat = coordinates[0][0]
        lon = coordinates[0][1]

        # Data creation
        data = [
            dict(
                type="scattergeo",
                lon=[lon],
                lat=[lat],
                mode='markers',
            )
        ]

        # Configuration of the zoom
        if lat is not None and lon is not None:
            lataxis = dict(range=[float(lat)-8.5, float(lat)+8.5])
            lonaxis = dict(range=[float(lon)-16, float(lon)+16])
        else:
            lataxis = None
            lonaxis = None

        layout = dict(
            geo=dict(
                lakecolor="rgb(255, 255, 255)",
                resolution=50,
                showcoastlines=True,
                showland=True,
                landcolor="rgb(229, 229, 229)",
                countrycolor="rgb(255, 255, 255)",
                coastlinecolor="rgb(255, 255, 255)",
                lataxis=lataxis,
                lonaxis=lonaxis
            ),
            margin=dict(l=10, r=10, t=0, b=0),
        )

        figure = dict(data=data, layout=layout)
        return figure

    @staticmethod
    def site_map_from_coordinates(lat, lon, text):
        """
        It returns a figure of a map with a scatter of the input information.
        """
        # Data creation
        data = [
            dict(
                type="scattergeo",
                lon=lon,
                lat=lat,
                text=text,
                mode='markers',
            )
        ]

        # Configuration of the zoom
        lataxis = dict(range=[min(lat)-9, max(lat)+9])
        lonaxis = dict(range=[min(lon)-15, max(lon)+15])

        layout = dict(
            geo=dict(
                lakecolor="rgb(255, 255, 255)",
                resolution=50,
                showcoastlines=True,
                showland=True,
                landcolor="rgb(229, 229, 229)",
                countrycolor="rgb(255, 255, 255)",
                coastlinecolor="rgb(255, 255, 255)",
                lataxis=lataxis,
                lonaxis=lonaxis
            ),
            margin=dict(l=10, r=10, t=0, b=0),
        )

        figure = dict(data=data, layout=layout)
        return figure

    @staticmethod
    def time_series_from_waterframe(waterframe, parameters=None):
        """
        It creates a Plotly time-series figure.

        Parameters
        ----------
            waterframe: WaterFrame
            parameters: str or list (optional, parameters=None)
                Parameters to plot.

        Returns
        -------
            figure: dict
                Plotly figure dictionary
        """

        figure = {'data': None, 'layout': None}

        # Check parameters
        if parameters is None:
            parameters = waterframe.parameters()
        elif isinstance(parameters, str):
            parameters = [parameters]
        elif isinstance(parameters, list):
            pass
        else:
            return figure

        data = [go.Scatter(x=waterframe.data.index, y=waterframe[parameter], fill="tozeroy",
                           name=parameter) for parameter in parameters]

        # Layout
        y_label = None
        if len(parameters) == 1:
            try:
                title = waterframe.meaning[parameters[0]]['long_name']
            except KeyError:
                title = parameters[0]
            try:
                y_label = waterframe.meaning[parameters[0]]['units']
            except KeyError:
                pass
        else:
            title = "_".join(parameters)
        min_value = None
        max_value = None
        for parameter in parameters:

            _min = min(waterframe[parameter])
            _max = max(waterframe[parameter])

            if min_value is None or _min < min_value:
                min_value = _min

            if max_value is None or _max > max_value:
                max_value = _max

        layout = {
            'title': title,
            'yaxis': {
                'range': [min_value, max_value],
                'title': y_label
            },
            'margin': {'l': 50, 'r': 1, 't': 45, 'b': 30}
        }

        figure = {"data": data, "layout": layout}

        return figure
